// Chris OG  \AMXtex\Export.cpp

#include "stdafx.h"
#include "Hxt.h"
#include "HashFile.h"
#include <atlbase.h>

#ifndef __IEMODEL_H
#include "IEModel.h"
#endif

#include "ExportModelDialog.h"
#include "ExportModelsDialog.h"
#include "ExportSingleAction.h"
#include "F_Protos.h"

#include "Resource.h"

#include "SDK\HModel.h"
#include "SDK\HAction.h"
#include "SDK\Misc.h"
#include "SDK\HProject.h"
#include "SDK\HChor.h"
#include "SDK\HProgres.h"
#include "SDK\HPropDri.h"

#include "ExportClasses.h"
#include "D3dx8math.h"
#include "Matrix34.h"

#ifndef __MATRIX44_H
#include "Matrix44.h"
#endif


#ifndef __VECTOR_H
#include "Vector.h"
#endif

#include <assert.h>
#include <math.h>
#include <float.h>
#include <string.h>

// Additional animation export headers
#include "Resource.h"

#ifndef __VECTOR2_H
#include "Vector2.h"
#endif

#include "Channel.h"
#include "Archive.h"

#ifndef __ROTATE_H
#include "Rotate.h"
#endif

// Begin Global Variables

CArray<VERTEXDATA, VERTEXDATA &>vertexlist;
CArray<POLYDATA, POLYDATA &>polylist;
CArray<BONEDATA, BONEDATA &>bonelist;
CArray<BONEVDILIST, BONEVDILIST &>vdilist;
CArray<MATERIALINFO, MATERIALINFO &>matlist;

CArray<RMPOLYMATINFO, RMPOLYMATINFO &>retainedpolylist;
CArray<int, int &>retainedmatlist;
CArray<DUPEVERT, DUPEVERT &>dupeverts;
CArray<int, int &>RMnewmats;


NORMALS* normallist;
POLYINFO* polynum;

int* matcount;
int newmatcount;
int unusedmats;

int bonenumber = 0;

int texture_map_id;

char remPathTextureName[128];

int totalvertcount;
int initialvertcount;

BOOL exportmaterials = TRUE;
BOOL exporttextures = TRUE;
BOOL exportnormals = TRUE;
BOOL exportVDIs = TRUE;
BOOL ingoreNormalVDIs = FALSE;
BOOL texturefacewhite = FALSE;
BOOL reversetexturesort = FALSE;
BOOL exportboneinfo = TRUE;
BOOL skinnedmeshexport = FALSE;
BOOL retainedmodeexport = FALSE;
float modelorientation = 1.0f; // 1=right-handed, -1=left-handed
BOOL exportWeightInfo = FALSE;

char floatstring[80];

Matrix34 BoneMatrix;
Matrix44 BoneSkinWeights;

DirectXType DXExportType;
ModelOrientation DXModelOrient;

// Animation Global variables
CArray<Channel2, Channel2 &>m_channellist;
CArray<FTMDATA, FTMDATA &>ftmlist;
CArray<ANIMDATA, ANIMDATA &>animlist;
Time    m_length;
int numchannels = 0;
Time totaltime2;
int fps = 30; // (hardcoded value)
int ftmlistpos = 0;
int ftmlistSize = 32;	// initial ftm list size

int frameunits;
String animationname;

ExpFileType FileExportType;

BOOL APOAppend = FALSE;
BOOL APONewCopy = FALSE;
BOOL APOReplace = FALSE;

TCHAR m_szPath[256];

// End Global Variables

BOOL FLOAT_EQ(float x, float v) {

	char f1[32];
	char f2[32];

	sprintf(f1, "%.6f",x);
	sprintf(f2, "%.6f",v);

	if (strcmp(f1, f2) == 0)
		return TRUE;
	else
		return FALSE;
}

BOOL CHxtApp::IsMemberOfArray(LONG *array, int arraylength, int member)
{
  for (int i=0; i<arraylength; i++) {
    if (array[i] == member)
      return TRUE;
  }

  return FALSE;
}

BOOL CHxtApp::OnExportModelCache(HModelCache *hmc)
{
   BOOL result;

   skinnedmeshexport = FALSE;
   retainedmodeexport = FALSE;

   CExportModelDialog cdlg(hmc->GetName());
   if (cdlg.DoModal()!=IDOK)
      return TRUE;
   CWaitCursor wait;
   m_iemp.m_isbuildmodelmap = cdlg.m_generatemodelmaps;
   m_iemp.m_minmodelmapsize = (MapLimit)cdlg.m_minsize;
   m_iemp.m_maxmodelmapsize = (MapLimit)cdlg.m_maxsize;
   m_iemp.m_subtype = (SubType)cdlg.m_subtype;
   m_iemp.m_isbuildweights = cdlg.m_expweightinfo;

   if (cdlg.m_expweightinfo) {		// if we're exporting weights, set the max weights
	   m_iemp.m_maxweightcount = 4;	// per vertex to 4 (DirectX limit)
	   //AfxMessageBox("Setting max weights to 4", MB_ICONINFORMATION);
   }

   exportmaterials = cdlg.m_expmatinfo;
   exporttextures = cdlg.m_exptexinfo;
   exportnormals = cdlg.m_expnorminfo;
   exportVDIs = cdlg.m_expvdis;
   ingoreNormalVDIs = cdlg.m_ignoreduplicatenormals;
   texturefacewhite = cdlg.m_settextowhite;
   reversetexturesort = cdlg.m_texturesort;
   exportboneinfo = cdlg.m_expboneinfo;

   DXExportType = (DirectXType)cdlg.m_exportstyle;

   if (DXExportType == DX_SM)
	   skinnedmeshexport = TRUE;
   else
	   retainedmodeexport = TRUE;

   DXModelOrient = (ModelOrientation)cdlg.m_modelorientation;

   if (DXModelOrient == DX_RIGHTHANDED)
	   modelorientation = 1.0f;
   else
	   modelorientation = -1.0f;

   if (cdlg.m_expweightinfo)
	   exportWeightInfo = true;
   else
	   exportWeightInfo = false;

   String modelmapfilename = cdlg.m_filename.GetWithoutSuffix() + ".bmp";
   IEPolyModel *model = hmc->BuildExportModel(&m_iemp, modelmapfilename.Get());
   if (model==NULL) {
      ReportExportError(cdlg.m_filename);
      return FALSE;
   }

   if (skinnedmeshexport && !retainedmodeexport)
	   result = ExportSkinnedMeshModel(model, cdlg.m_filename);
   else
	   result = ExportRetainedModeModel(model, cdlg.m_filename);

   DeleteIEModel(model);

   if (!result)
      ReportExportError(cdlg.m_filename);

   return result;
}

BOOL CHxtApp::OnExportActionCache(HActionCache *hac)
{
	// Clear Globals, in case it's being reused
	bonenumber = 0;
	m_length.SetTicks(0L);
	numchannels = 0;
	totaltime2.SetTicks(0L);
	fps = 30; // (hardcoded value)
	ftmlistpos = 0;
	skinnedmeshexport = FALSE;
	retainedmodeexport = FALSE;
	modelorientation = 1.0f; // 1=right-handed, -1=left-handed
	// done reseting globals

   BOOL GotValidFile = FALSE;

   HModelCache *hmc = hac->GetDefaultModelCache();

   if (hmc==NULL) {
      CString msg, title;
      msg.LoadString(IDS_ACTIONHASNOMODELASSIGNED);
      String temp = (LPCTSTR)msg;
      temp.Replace("%1", hac->GetName());
      msg = temp.Get();
      title.LoadString(IDS_ANIMEXPORTTITLE);

      MessageBox(GetMainApplicationWnd(), msg, title, MB_OK);
      return FALSE;
   }

   OPENFILENAME ofn;
   memset( &ofn, 0, sizeof(ofn) );
   static TCHAR file[256];
   static TCHAR szFilepath[256];
   static TCHAR fileTitle[256];
   static TCHAR filter[] =
      TEXT("X files (*.x)\0*.x\0")
      TEXT("All Files (*.*)\0*.*\0");
   _tcscpy( file, TEXT(""));
   _tcscpy( fileTitle, TEXT(""));
                    
   ofn.lStructSize       = sizeof(ofn);
   ofn.hwndOwner         = GetMainApplicationWnd();
   ofn.hInstance         = NULL;//m_hInstance;
   ofn.lpstrFilter       = filter;
   ofn.lpstrCustomFilter = NULL;
   ofn.nMaxCustFilter    = 0L;
   ofn.nFilterIndex      = 1L;
   ofn.lpstrFile         = file;
   ofn.nMaxFile          = sizeof(file);
   ofn.lpstrFileTitle    = fileTitle;
   ofn.nMaxFileTitle     = sizeof(fileTitle);
   ofn.lpstrInitialDir   = NULL;
   ofn.nFileOffset       = 0;
   ofn.nFileExtension    = 0;
   ofn.lpstrDefExt       = TEXT("*.x");
   ofn.lCustData         = 0;
                    
   ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
   if ( ! GetOpenFileName( &ofn) )
   {
	   TCHAR s[40];
       DWORD dwErr = CommDlgExtendedError();
       if ( 0 != dwErr )
       {
		   wsprintf( s, "GetOpenFileName failed with %x", dwErr );
		   MessageBox( GetMainApplicationWnd(), s, "TexWin", MB_OK | MB_SYSTEMMODAL );
	   }
	   return 0;
   }

   lstrcpy(m_szPath, ofn.lpstrFile);

   ExportSingleAction dlg(hac);

   if (dlg.DoModal()!=IDOK)
	   return TRUE;

   frameunits = dlg.m_animtiming;
   animationname = dlg.m_animationsetname;

   CString tempAnimName = animationname.Get();
   tempAnimName = GetValidDirectXName(tempAnimName);
   animationname = tempAnimName;

   FileExportType = (ExpFileType)dlg.m_fileexptype;

   if (FileExportType == EFT_NEWCOPY) {
	   APONewCopy = TRUE;
   	   APOAppend = FALSE;
   }
   else {
	   APOAppend = TRUE;
   	   APONewCopy = FALSE;
   }

   DXExportType = (DirectXType)dlg.m_directxtype;

   if (DXExportType == DX_SM)
	   skinnedmeshexport = TRUE;
   else
	   retainedmodeexport = TRUE;

   float errortolerance = 1.0f;

   if (skinnedmeshexport)
	   return ExportSMAnimationSet(errortolerance, hac, hmc);
   
   if (retainedmodeexport)
	   return ExportRMAnimationSet(errortolerance, hac, hmc);

   return TRUE;
}

BOOL CHxtApp::OnExportObjectCacheContainer(HObjectCacheContainer *hocc)
{
   BOOL result;

   // Set defaults
   skinnedmeshexport = FALSE;
   retainedmodeexport = FALSE;

   CExportModelsDialog cdlg(hocc);
   if (cdlg.DoModal()!=IDOK)
      return TRUE;

   CWaitCursor wait;
   m_iemp.m_isbuildmodelmap = cdlg.m_generatemodelmaps;
   m_iemp.m_minmodelmapsize = (MapLimit)cdlg.m_minsize;
   m_iemp.m_maxmodelmapsize = (MapLimit)cdlg.m_maxsize;
   m_iemp.m_subtype = (SubType)cdlg.m_subtype;
   m_iemp.m_isbuildweights = cdlg.m_expweightinfo;

   if (cdlg.m_expweightinfo) {		// if we're exporting weights, set the max weights
	   m_iemp.m_maxweightcount = 4;	// per vertex to 4 (DirectX limit)
	   //AfxMessageBox("Setting max weights to 4", MB_ICONINFORMATION);
   }



   exportmaterials = cdlg.m_expmatinfo;
   exporttextures = cdlg.m_exptexinfo;
   exportnormals = cdlg.m_expnorminfo;
   exportVDIs = cdlg.m_expvdis;
   ingoreNormalVDIs = cdlg.m_ignoreduplicatenormals;
   texturefacewhite = cdlg.m_settextowhite;
   reversetexturesort = cdlg.m_texturesort;
   exportboneinfo = cdlg.m_expboneinfo;

   DXExportType = (DirectXType)cdlg.m_exportstyle;

   if (DXExportType == DX_SM)
	   skinnedmeshexport = TRUE;
   else
	   retainedmodeexport = TRUE;

   DXModelOrient = (ModelOrientation)cdlg.m_modelorientation;

   if (DXModelOrient == DX_RIGHTHANDED)
	   modelorientation = 1.0f;
   else
	   modelorientation = -1.0f;

   if (cdlg.m_expweightinfo)
	   exportWeightInfo = true;
   else
	   exportWeightInfo = false;

   for (POSITION pos = cdlg.m_selected.GetHeadPosition(); pos; ) {
      HModelCache *hmc = cdlg.m_selected.GetNext(pos);
      String filename = hmc->GetName();
      filename = filename.GetNameOnlyWithoutSuffix() + ".X";
      filename = CombineNameToPath(cdlg.m_dir, filename);
      String modelmapfilename = filename.GetWithoutSuffix() + ".bmp";
      IEPolyModel *model = hmc->BuildExportModel(&m_iemp, modelmapfilename.Get());
      if (model==NULL) {
         ReportExportError(filename);
         return FALSE;
      }

	  if (skinnedmeshexport && !retainedmodeexport)
		  result = ExportSkinnedMeshModel(model, filename);
	  else
		  result = ExportRetainedModeModel(model, filename);

      DeleteIEModel(model);
      if (!result) {
         ReportExportError(filename);
         return FALSE;
      }
   }
   return TRUE;
}

void CHxtApp::ReportExportError(const String &filename)
{
   CString caption;
   caption.LoadString(IDS_EXPORTTITLE);
   CString text;
   text.LoadString(IDS_ERROREXPORTINGMODEL);
   String temp = (LPCTSTR)text;
   temp.Replace("%1", filename);
   text = temp.Get();
   MessageBox(GetMainApplicationWnd(), text, caption, MB_OK); 
}
 
static void CountBoneBranch(IEBone *bone, LONG *count)
{
   for (;bone;bone=bone->m_sibling) {
      (*count)++;
      if (bone->m_child)
         CountBoneBranch(bone->m_child, count);
   }  
}

LONG CHxtApp::CountBones(IEPolyModel *model)
{
   LONG count=0;
   CountBoneBranch(model->m_bonehead, &count);

   return count;
}


BOOL CHxtApp::ExportSkinnedMeshModel(IEPolyModel *model, const String &filename)
{
   FileStream fs; //cjr
   LONG totalitems, curritem = 0;

   newmatcount = 0;
   unusedmats = 0;
   texture_map_id = 0;
   totalvertcount = 0;
   bonenumber = 0;

   polynum = new POLYINFO[model->m_polyarray.GetSize()];

   int newpolynum=model->m_polyarray.GetSize();
   
   InitNewPolys(newpolynum);

   InitMatList(model);

   m_hprogressbar = GetProgressBar();
   if (m_hprogressbar) {
      m_hprogressbar->Initialize();
      CString text;
      text.LoadString(IDS_EXPORTTITLE);
      m_hprogressbar->SetMessage1((LPCTSTR)text);
      m_hprogressbar->SetMessage2((LPCTSTR)filename.GetNameOnly().Get());
   }
 
   // Get total items for percent bar
   totalitems = model->m_attrlist.GetSize() +
	   model->m_maplist.GetSize() +
	   model->m_vertexlist.GetSize() +
	   model->m_normallist.GetSize() +
	   model->m_polyarray.GetSize() +
	   CountBones(model);

   initialvertcount = model->m_vertexlist.GetSize();

   if (!fs.Open( (String)filename, HashStream::HS_STORE))
      return FALSE;

   if(exportnormals && exportVDIs)
	   GetVertexPolyNormalData(fs, model, totalitems, curritem);
   else if (!exportnormals && exportVDIs)
	   GetVertexPolyData(fs, model, totalitems, curritem);
   else if (!exportVDIs)
	   GetVertexPolyNormalDataNoVDIs(fs, model, totalitems, curritem);

   fs << "xof 0303txt 0032" << nl << nl;

   //If needed, write the VDI template to the file
   if(totalvertcount > model->m_vertexlist.GetSize())
	   WriteVDIHeader(fs);

   if (exportboneinfo) {
	   WriteBoneHeader(fs);
	   SaveBones(fs, model, totalitems,curritem);
   }

   // Calculate the texture stuff

   if (exportmaterials && (model->m_attrlist.GetSize() > 0))
	   MaterialCleanup(fs, model);

   SaveVertexList(fs, model, totalitems, curritem);
   SavePolys(fs, model, totalitems, curritem);

   if (exportnormals) {
	   SaveNormalList(fs, model, totalitems, curritem);
	   SavePolys(fs, model, totalitems, curritem);
	   fs << "}" << nl; //needed to close the normal list
   }


   if (exportmaterials && (model->m_attrlist.GetSize() > 0)) {
	   SaveAttributeIDList(fs, model, totalitems, curritem);

	   SaveAttributes(fs, model, totalitems, curritem);
   }

   if (exporttextures)
	   WriteUVList(fs, model);

   if(exportVDIs && (totalvertcount > model->m_vertexlist.GetSize()))
      WriteVDIs(fs, model, totalitems, curritem);

   if (exportboneinfo)
	   SaveSkinInfo(fs, model, totalitems, curritem);

   fs << "}                   // End Mesh" << nl;
   if (exportboneinfo)
	   fs << "}                   // End RootFrame" << nl;

   delete [] polynum;
   delete [] matcount;

   if (m_hprogressbar) {
      m_hprogressbar->Uninitialize();
      m_hprogressbar = NULL;
   }

   // Clear the vertexlist and polylist CArrays

//   for (int i=vertexlist.GetSize()-1;i>=0;i--) {
//	   if(vertexlist.GetAt(i) != NULL)
//		   delete [] vertexlist.GetAt(i) ; 
	   //VERTEXDATA tmp = vertexlist.GetAt(i);
	   //delete tmp.modified;
//   }

   vertexlist.SetSize(0);
   vertexlist.RemoveAll();
   vertexlist.FreeExtra();
   polylist.SetSize(0);
   polylist.RemoveAll();
   polylist.FreeExtra();
   bonelist.SetSize(0);
   bonelist.RemoveAll();
   bonelist.FreeExtra();
   matlist.SetSize(0);
   matlist.RemoveAll();
   matlist.FreeExtra();

   return TRUE;
}

BOOL CHxtApp::ValidPolyMap(IEPolyModel *model, int polynum)
{
	BOOL use_color = FALSE;
	int color_id   = -1;
	int other_id    = -1;
	int mapcount;
	
	IEPoly *poly = model->m_polyarray[polynum];
	mapcount = poly->m_mapnodearray.GetSize();

	if (mapcount == 0) // if there are no maps on this poly
		return FALSE;

	if (!reversetexturesort)	{  // sort normal way (top->bottom)
		for (int i=mapcount-1; i>=0; i--) {
			IEMapNode *mapnode = poly->m_mapnodearray[i];
			switch (mapnode->m_kind) {
				case IEMapNode::COLORMAP:
					if (color_id == -1) {
					color_id=i;
					use_color=TRUE;
					}
					break;
				case IEMapNode::BUMPMAP:
					if (other_id == -1)
						other_id = i;
					break;
				case IEMapNode::DIFFUSEMAP:
					if (other_id == -1)
						other_id = i;
					break;
				case IEMapNode::COOKIECUTMAP:
					if (other_id == -1)
						other_id = i;
					break;
				default:
					break;
			}
		}
	}
	else {	// sort reverse (bottom->top)
		for (int i=0; i<=mapcount-1; i++) {
			IEMapNode *mapnode = poly->m_mapnodearray[i];
			switch (mapnode->m_kind) {
				case IEMapNode::COLORMAP:
					if (color_id == -1) {
					color_id=i;
					use_color=TRUE;
					}
					break;
				case IEMapNode::BUMPMAP:
					if (other_id == -1)
						other_id = i;
					break;
				case IEMapNode::DIFFUSEMAP:
					if (other_id == -1)
						other_id = i;
					break;
				case IEMapNode::COOKIECUTMAP:
					if (other_id == -1)
						other_id = i;
					break;
				default:
					break;
			}
		}
	}

	if (use_color) {  // if you can use a color map, do it
		return TRUE;
	}

	else {
		if (other_id != -1) { // if no color, use bump if it has one
			return TRUE;
		}
	}

	return FALSE;  // No good ID, return false;
}


char* CHxtApp::RemovePathname(FileStream &fs, char* fullpathname)
{
	char texturename[128];
	int pathposition;
	int pathnamelength;
	int x;
	int strpos = 0;
	char *pdest;

	pathnamelength=strlen(fullpathname);


	pdest = strrchr( fullpathname, '\\' );
	pathposition = pdest - fullpathname + 1;

	for (x=pathposition;x<pathnamelength;x++) {
		texturename[strpos]=fullpathname[x];
		strpos++;
	}
	
	//add the terminating character
	texturename[strpos]='\0';

	// this is to prevent 4172 error (returning a local var)
	strcpy(remPathTextureName,texturename);
	return remPathTextureName;
}

void CHxtApp::InitNewPolys(int newpolynum)
{
	for (int x=0; x<newpolynum; x++)
	{
		polynum[x].modified=0;
		polynum[x].material=0;
	}
}

void CHxtApp::MaterialCleanup(FileStream &fs, IEPolyModel *model)
{
	int poly_count = model->m_polyarray.GetSize();

	int orig_attr_count = model->m_attrlist.GetSize();
	int total_mats = orig_attr_count+newmatcount;
	int w,x,y,z;

	IEPoly **polyarray = model->m_polyarray.GetArray();

	// allocate space to the matcount array
	matcount = new int[matlist.GetSize()];

	IEMap *maparray = model->m_maplist.GetArray();

	// First, init the matcount array elements to 0
	for (w=0;w<total_mats;w++)
		matcount[w]=0;

	// Second, cycle through the material list to weed out
	// unused materials.  To do this, set the matcount value to
	// 1 if it's used.  If at the end it's 0, then it's unused

	for (x=0;x<poly_count;x++)
	{
		matcount[polylist[x].matlistindex] = 1;
	}

	// Finally, use the matlist to delete unused materials and
	// shift the material ids of the remaining materials
	// Note: Only loop through the original mats

	// we need to now remove the unused materials, but there is now only
	// one list -- the matlist.  No more attr list and newmat list

	total_mats = matlist.GetSize();

	for (y=(total_mats-1);y>=0;y--) {
		if (matcount[y] == 0) {
			// delete the material

			matlist.RemoveAt(y, 1);

			// now lower the ids of all the materials higher than this one
			for(z=0;z<poly_count;z++) {
				if(polylist[z].matlistindex > y) {
					polylist[z].matlistindex--;
				}
			}

		}
	}

	// now that the unused mats have been removed,
	// consolidate any duplicates
	for (y=0; y<matlist.GetSize();y++) {
		for (x=0;x<matlist.GetSize();x++) {
			if (x != y) { // don't compare a material to itself

				BOOL match = TRUE;

				if (!FLOAT_EQ(matlist[y].red,matlist[x].red))
					match = FALSE;

				if (!FLOAT_EQ(matlist[y].green,matlist[x].green))
					match = FALSE;

				if (!FLOAT_EQ(matlist[y].blue,matlist[x].blue))
					match = FALSE;

				if (!FLOAT_EQ(matlist[y].alpha,matlist[x].alpha))
					match = FALSE;

				if (!FLOAT_EQ(matlist[y].specular,matlist[x].specular))
					match = FALSE;

				CString map1 = maparray[matlist[y].texture].m_filename;
				CString map2 = maparray[matlist[x].texture].m_filename;

				if (strcmp(map1, map2) != 0)
					match = FALSE;

				if (match) {
					// it's a match, so remove the higher number and set it to the lower one
					// delete the material and lower the higher ids
					if (x > y) {
						matlist.RemoveAt(x, 1);

						for(z=0;z<poly_count;z++) {
							// if it's the higher mat number, change to the lower one
							if(polylist[z].matlistindex == x)
								polylist[z].matlistindex = y;

							// lower the ids of all the materials higher than this one
							if(polylist[z].matlistindex > x)
								polylist[z].matlistindex--;
						}
					}
					else { // y > x
						matlist.RemoveAt(y, 1);

						// now lower the ids of all the materials higher than this one
						for(z=0;z<poly_count;z++) {
							// if it's the higher mat number, change to the lower one
							if(polylist[z].matlistindex == y)
								polylist[z].matlistindex = x;

							// lower the ids of all the materials higher than this one
							if(polylist[z].matlistindex > y)
								polylist[z].matlistindex--;
						}
					}
				}
			}
		}
	}
}

void CHxtApp::SaveAttributes(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
   int count = model->m_attrlist.GetSize();
   IEMap *maparray = model->m_maplist.GetArray();
   char texturename[128];

   for (int i=0; i<matlist.GetSize();i++) {
	  fs << nl << "Material {          // Material" << nl;
	  sprintf(floatstring, "%.6f;%.6f;%.6f;%.6f;;",
		  matlist[i].red,
		  matlist[i].green,
		  matlist[i].blue,
		  matlist[i].alpha);

	  fs  << floatstring << nl;

	  if (matlist[i].texture == -1) {
		  if(FLOAT_EQ(0.000000,matlist[i].specular)) {
			  fs << "0.000000;" << nl;
			  fs << "0.000000;0.000000;0.000000;;" << nl;
			  fs << "0.000000;0.000000;0.000000;;" << nl;
			  fs << "} // End Material" << nl;

		  }
		  else {
			  fs << "10.000000;" << nl;
			  sprintf(floatstring, "%.6f;%.6f;%.6f;;",
				  matlist[i].specular,
				  matlist[i].specular,
				  matlist[i].specular);

			  fs  << floatstring << nl;
		  
			  fs << "0.000000;0.000000;0.000000;;" << nl;

			  fs << "} // End Material" << nl;
		  }
	  }
	  else { // it's a textured material
		   fs << "299.000000;" << nl;

		   fs << "0.000000;0.000000;0.000000;;" << nl;
		   fs << "0.000000;0.000000;0.000000;;" << nl;

		   strcpy(texturename, RemovePathname(fs, maparray[matlist[i].texture].m_filename));
		   fs << "TextureFilename {\"" << texturename << "\";}" << nl;
		   fs << "} // End Material" << nl;
	  }
    }

   fs << "} // End MeshMaterialList" << nl;
}


void CHxtApp::SaveVertexList(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
   int count = totalvertcount;

   fs << "Mesh {" << nl;
   fs << count << ";                  // Number of Vertices" << nl;

   for (int i=0; i < count; i++) {
	   if (m_hprogressbar)
		   m_hprogressbar->SetProgress((int)((float)curritem++/totalitems*100));

   	   sprintf(floatstring, "%.6f;%.6f;%.6f;",
		   vertexlist[i].x, vertexlist[i].y, vertexlist[i].z);//*modelorientation);
	   
	  fs  << floatstring;

	  if (i+1 < count)         // if this is NOT the last vertex,
		  fs << "," << nl; // add a coma
   }

   fs << ";" << nl; // Add the vertex-closing semicolon
}

void CHxtApp::SaveAttributeIDList(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
   int attr_count;
   int poly_count = model->m_polyarray.GetSize();

   if (exporttextures)
	   attr_count = (model->m_attrlist.GetSize()+newmatcount)-unusedmats;
   else
	   attr_count = model->m_attrlist.GetSize();

   IEPoly **polyarray = model->m_polyarray.GetArray();

   fs << nl << "MeshMaterialList {              //Material List" << nl;
   fs << matlist.GetSize() << ";" << nl;
   fs << poly_count << ";" << nl;


   for (int i=0; i < poly_count; i++) {
	   if (m_hprogressbar)
		   m_hprogressbar->SetProgress((int)((float)curritem++/totalitems*100));

	   fs << polylist[i].matlistindex;

	   if (i+1 < poly_count)         // if this is NOT the last vertex point,
		   fs << "," << nl;  // add a coma
   }

   fs << ";" << nl;  // Add the vertex-closing semicolon
}


void CHxtApp::SaveNormalList(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
   int count = totalvertcount;

   fs << nl << "MeshNormals {" << nl;

   fs << count << ";" << nl;

   for (int i=0; i < count; i++) {
	   sprintf(floatstring, "%.6f;%.6f;%.6f;",
		   vertexlist[i].nx, vertexlist[i].ny, vertexlist[i].nz);//*modelorientation);
	   fs  << floatstring;

	   if (i+1 < count)          // if this is NOT the last vertex,
		   fs << "," << nl;
   }

   fs << ";" << nl;

   delete [] normallist;
}


void CHxtApp::SavePolys(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
   int count = model->m_polyarray.GetSize();

   fs << count << ";                  // Number of Polys" << nl;

   for (int i=0; i < count; i++) {
	   if (m_hprogressbar)
		   m_hprogressbar->SetProgress((int)((float)curritem++/totalitems*100));

	   if (FLOAT_EQ(modelorientation, 1.0)) // if it's right handed
		   fs  << "4;"
		       << polylist[i].vertex[0] << ","
			   << polylist[i].vertex[1] << ","
			   << polylist[i].vertex[2] << ","
			   << polylist[i].vertex[3] << ";";
	   else	// it's left handed, so reverse the poly's vertex order
		   fs  << "4;"
		       << polylist[i].vertex[3] << ","
			   << polylist[i].vertex[2] << ","
			   << polylist[i].vertex[1] << ","
			   << polylist[i].vertex[0] << ";";

	   if (i+1 < count)         // if this is NOT the last vertex,
		   fs << "," << nl;	// add a coma
   }

   fs << ";" << nl; // Add the poly-closing semicolon
}

void ComputeBoneToStandard(Vector m_pivot, Vector m_end, Vector m_rollhandle)
{
	BoneMatrix = Rotate2V((m_end-m_pivot), Vector( 0, 0, 1 ));
	Vector rollvec = BoneMatrix * (m_rollhandle-m_pivot);
	float roll = (float)atan2( -rollvec.x, rollvec.y );
	if ( roll )
		BoneMatrix = ZRotateRadians( -roll ) * BoneMatrix;
	BoneMatrix.SetTranslate( BoneMatrix * -m_pivot );
}

Matrix44 CHxtApp::GetMatrixConversion(Matrix44 mOrig, Matrix44 mFinal)
{
	// mRot = mFinal . mOrig^-1
	Matrix44 mRot;

	mRot = mFinal * mOrig.Inverse();

	return mRot;
}

void CHxtApp::SaveBoneBranch(FileStream &fs, IEBone *bone, String &spaces, Matrix44 Parent, Matrix44 CombinedParent, LONG &totalitems, LONG &curritem)
{
	Matrix44 BoneTransform;
	Matrix44 NewCombinedTransform;
	Vector4 v;
	char vstring[80];
	Vector4 v1, v2, v3, v4;
	Matrix44 final, final2;
	Matrix44 xrot, yrot, zrot;
	int x;

	Matrix44 AbsoluteBoneMat;

	// transposed rotations
	v1.Set(1.000000,0.000000,0.000000,0.000000);
	v2.Set(0.000000,0.000000,-1.000000,0.000000);
	v3.Set(0.000000,1.000000,0.000000,0.000000);
	v4.Set(0.000000,0.000000,0.000000,1.000000);
	xrot = Matrix44(v1, v2, v3, v4);

	v1.Set(0.000000,0.000000,1.000000,0.000000);
	v2.Set(0.000000,1.000000,0.000000,0.000000);
	v3.Set(-1.000000,0.000000,0.000000,0.000000);
	v4.Set(0.000000,0.000000,0.000000,1.000000);
	yrot = Matrix44(v1, v2, v3, v4);

	v1.Set(0.000000,-1.000000,0.000000,0.000000);
	v2.Set(1.000000,0.000000,0.000000,0.000000);
	v3.Set(0.000000,0.000000,1.000000,0.000000);
	v4.Set(0.000000,0.000000,0.000000,1.000000);
	zrot = Matrix44(v1, v2, v3, v4);

	for (; bone; bone = bone->m_sibling) {

	  ComputeBoneToStandard(bone->m_pivotpos, bone->m_endpos, bone->m_rollpos);

	  AbsoluteBoneMat = ToMatrix44 ( BoneMatrix );

	  // Set BoneTransform to GetMatrixConversion(Matrix44 mOrig, Matrix44 mFinal)
	  BoneTransform = GetMatrixConversion(Parent, AbsoluteBoneMat);

	  BoneTransform = BoneTransform.Transpose();
	  BoneTransform = BoneTransform.Inverse();  // new 10/29 from import/export help file

	  // Populate the Bonelist skinweight values
	  // first get a valid directx name
	  CString tempBoneName = bone->m_name;
	  tempBoneName = GetValidDirectXName(tempBoneName);
	  //strcpy(bonelist[bonenumber].name, bone->m_name);
	  strcpy(bonelist[bonenumber].name, tempBoneName);

	  // combined parent
	  
	  NewCombinedTransform =  BoneTransform * CombinedParent;

	  bonelist[bonenumber].SkinWeightOffset = NewCombinedTransform.Inverse();

	  // Output the frame info
  	  fs << spaces << "Frame ";

	  CString BoneName = bone->m_name;
	  BoneName = GetValidDirectXName(BoneName);
	  fs.Write( BoneName, BoneName.GetLength());

	  fs << " {" << nl;

  	  fs << spaces << "  FrameTransformMatrix {" << nl;

	  for (x=0;x<4;x++) {
		  v = BoneTransform[x];

		  sprintf(vstring, "    %.6f,%.6f,%.6f,%.6f", v[0],v[1],v[2],v[3]);
	  	  fs << spaces << vstring;
		  if (x < 3)
			  fs << "," << nl;
		  else
			  fs << ";;" << nl << spaces << "  }" << nl;
	  }

	  //Increase the bonenumber for use in the bonelist
	  bonenumber++;
      
      if (bone->m_child) {
		  String childspaces = spaces + "  ";
		  SaveBoneBranch( fs, bone->m_child, childspaces, AbsoluteBoneMat, NewCombinedTransform, totalitems, curritem );
	  }

	  if (spaces.length() > 0)		 // if it's not the root frame,
		  fs << spaces << "}" << nl; // close the frame
   }
}

void CHxtApp::SaveBones(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
	Vector4 v1,v2, v3, v4;

	LONG bc = CountBones(model);

	bonelist.SetSize(bc); // list of bones and their SkinWeightOffsets


	Matrix44 offset;
	String spaces = "";

	offset.SetIdentity(); // set initial bone to identity matrix

	if (modelorientation < 0) {
		v1 = offset[0];
		v2 = offset[1];
		v3 = offset[2];
		v4 = offset[3];

		v3[2] = -1.0f;
		
		offset = Matrix44(v1, v2, v3, v4);
	}


	// only use offset in both places for testing purposes
	// change it to another id matrix later on,
	// (when root offset isn't neccesarily the id matrix)
	SaveBoneBranch(fs, model->m_bonehead, spaces, offset, offset, totalitems, curritem);

	// Now that we've stored all the bones, check if duplicate bone names
	// have been created due to the valid DirectX name conversion
	DuplicateNameCheck();

}

void CHxtApp::SaveSkinInfo(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
	fs << nl;
	fs << "XSkinMeshHeader {" << nl;
	fs << "  2;" << nl << "  4;" << nl;
	fs << "  " << bonenumber << ";" << nl;
	fs << "}" << nl << nl;

	SaveSkinWeights(fs, model->m_bonehead, totalitems, curritem);
}


void CHxtApp::SaveSkinWeights(FileStream &fs, IEBone *bone, LONG &totalitems, LONG &curritem)
{
	int boneidcount;
	int i;
	Matrix44 MatrixOffset;
	Vector4 v;
	char vstring[80];
	String bonename;
	BONEVDILIST newvdi;
	int newboneidcount;
	int weightCount;

	int vdistart = initialvertcount;
	int vdiend = totalvertcount; // is this off by one perhaps?
	int currentpos, endpos;

	endpos = vdiend;

	for (; bone; bone = bone->m_sibling) {
		// for loop this for all bones
		fs << "SkinWeights {" << nl;
		fs << "  \"";

		CString BoneName = bone->m_name;
		BoneName = GetValidDirectXName(BoneName);
		fs.Write( BoneName, BoneName.GetLength());

		fs << "\";" << nl;

		boneidcount = bone->m_vertexidarray.GetSize();

		LONG *array = bone->m_vertexidarray.GetArray();

		weightCount = bone->m_weightarray.GetSize();
		float *weightArray = bone->m_weightarray.GetArray();

		//char debugMsg[256];
		//sprintf(debugMsg,"Weight Array for bone %s has %d entries",
		//	BoneName, weightCount);
		//AfxMessageBox(debugMsg, MB_ICONINFORMATION);

		//for (int j=0;j<weightCount;j++) {
		//	sprintf(debugMsg,"Weight for id %d (vert %d): %.6f",
		//		j, array[j], weightArray[j]);
		//	AfxMessageBox(debugMsg, MB_ICONINFORMATION);
		//}

		// cycle through all the vdis
		for (currentpos=vdistart; currentpos<endpos; currentpos++) {
			if (IsMemberOfArray(array, boneidcount, vertexlist[currentpos].originalvertex)) { // original vertex is in this bone's vertex array
				newvdi.v = currentpos;
				// get VDI weight (based off original vertex) if exportWeightInfo is true
				if (exportWeightInfo) {
					bool foundWeight = false;
					// loop through cps for this bone to find the index.  Once found,
					// store the corresponding value from the weight array
					// (this should ALWAYS return a valid weight, so through error if not
					for (int w=0; w<weightCount; w++) {
						if (foundWeight) // we found our weight, so break out of loop
							break;

						if (array[w] == vertexlist[currentpos].originalvertex) {
							newvdi.w = weightArray[w];
							foundWeight = true;
						}
					}
					// if we got here (looped through all cp's) and did not find a weight
					// throw an error
					if (!foundWeight) {
						char debugMsg[256];
						sprintf(debugMsg, "ERROR: Could not find weight for vertex %d in bone %s",
							vertexlist[currentpos].originalvertex,
							BoneName);
						AfxMessageBox(debugMsg, MB_ICONINFORMATION);
					}
				}
				vdilist.Add(newvdi);
			}
		}

		newboneidcount = boneidcount + vdilist.GetSize();

		// Now write the SkinWeights list
		// -int newboneidcount = boneidcount+ vdilist.GetSize();
		// -not sure if VDIs go first or last (does it matter?)
		// -make sure to add extra entries both for vertexid AND skin weight

		fs << "  " << newboneidcount << ";";

		if (newboneidcount == 0) // just output semicolons and nothing else
			fs << ";;" << nl;
		else { // output the skinweight data
			fs << nl;

			// Output the original vertex id list
			for (i=0; i < boneidcount; i++) {
				fs << "  " << array[i];

				if (i < (boneidcount-1))
					fs << "," << nl;
			}

			// need to fix this for bones with boneidcount==0 - see empty skin weights
			if ((vdilist.GetSize() == 0) && (boneidcount > 0))
				fs << ";" << nl;
			else
				fs << "," << nl;
			
			// Output the VDI vertex id list
			for (i=0; i < vdilist.GetSize(); i++) {
				fs << "  " << vdilist[i].v;
				if (i < (vdilist.GetSize()-1))
					fs << "," << nl;
			}

			if (vdilist.GetSize() > 0)
				fs << ";" << nl;

			// output the skin weight list
			if (exportWeightInfo) {	// export proper weight info for each vertex
				char weightInfo[64];
				// output the original weight list
				for (i=0; i < weightCount; i++) {
					sprintf(weightInfo, "  %.6f", weightArray[i]);
					fs << weightInfo;

					if (i < (weightCount-1))
						fs << "," << nl;
				}

				if ((vdilist.GetSize() == 0) && (weightCount > 0))
					fs << ";" << nl;
				else
					fs << "," << nl;

				// output the VDI weight list
				for (i=0; i < vdilist.GetSize(); i++) {
					sprintf(weightInfo, "  %.6f", vdilist[i].w);
					fs << weightInfo;
					if (i < (vdilist.GetSize()-1))
						fs << "," << nl;
				}

				if (vdilist.GetSize() > 0)
					fs << ";" << nl;
			}
			else { // export hardcoded weights of 1.000000 for each vertex
				for (i=0; i < newboneidcount; i++) {
					fs << "  1.000000";

					if (i < (newboneidcount-1))
						fs << "," << nl;
					else
						fs << ";" << nl;
				}
			}
		}

		// output the matrix offset

		// clear out the current list, to get ready for the nextbone
		vdilist.SetSize(0);
		vdilist.RemoveAll();
		vdilist.FreeExtra();

		// Convert name to valid directx name (need to convert to CString first)
		CString tempBoneName = bone->m_name;
		tempBoneName = GetValidDirectXName(tempBoneName);
		bonename = tempBoneName;

		MatrixOffset = GetSkinWeightOffset(bonename);

		for (i=0;i<4;i++) {
			v = MatrixOffset[i];
			sprintf(vstring, "  %.6f,%.6f,%.6f,%.6f", v[0],v[1],v[2],v[3]);
			fs << vstring;
			if (i < 3)
				fs << "," << nl;
			else
				fs << ";;" << nl << "}" << nl << nl;
		}
		
		if (bone->m_child)      
			SaveSkinWeights( fs, bone->m_child, totalitems, curritem );
	}
}
   

Matrix44 CHxtApp::GetSkinWeightOffset(String &bonename)
{
	BOOL foundbone=FALSE;
	int count = 0;

	while ((!foundbone) && (count < bonenumber))
	{
		if (bonename == bonelist[count].name)
			foundbone=TRUE;
		else
			count++;
	}

	if (foundbone == FALSE) {
		char debugMsg[256];
		sprintf(debugMsg,"ERROR: Could not find SkinWeightOffset for Bone: %s (Count = %d)",bonename, count);
		AfxMessageBox(debugMsg, MB_ICONINFORMATION);
	}

	return bonelist[count].SkinWeightOffset;
}

  
void CHxtApp::WriteUVList(FileStream &fs, IEPolyModel *model)
{
   int vertsize = totalvertcount;

   vertsize = min(65536, vertsize);

   fs << nl << "MeshTextureCoords {" << nl;

   fs << vertsize << ";" << nl;

   for (int i=0; i < vertsize; i++) {
	   sprintf(floatstring, "%.6f;%.6f;", vertexlist[i].u, vertexlist[i].v);
	   fs  << floatstring;

	   if (i+1 < vertsize)          // if this is NOT the last vertex,
		   fs << "," << nl;
   }

   fs << ";" << nl << "}" << nl;
}

void CHxtApp::GetVertexPolyNormalData(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
	VERTEXDATA verttoadd;

	BOOL uvsexist=FALSE;	// For faster code, these are 
	BOOL uvsmatch=FALSE;	// calculated only once per loop
	BOOL normalsmatch=FALSE;

	BOOL foundmatch = FALSE;  // used in for loop

	int j, k; // loop counters

	// Get the various initial CArray sizes
	int vertcount = model->m_vertexlist.GetSize();
	totalvertcount = vertcount;
	int polycount = model->m_polyarray.GetSize();

	Vector *normalarray = model->m_normallist.GetArray();
	Vector *vertexarray = model->m_vertexlist.GetArray();

	// Set the CArray sizes
	vertexlist.SetSize(vertcount); // This one will likely grow
	polylist.SetSize(polycount);

	vertcount = min(65536, vertcount);

	BOOL vpmap = FALSE;

	// Initialize the vertexlist modified count
	// Only init these values to save time, as the program
	// will never check other fields if modified=0;

	for (int i=0; i < vertcount; i++) {
		vertexlist[i].modified=0;
		vertexlist[i].u=0.0f;  // if all situations write this,
		vertexlist[i].v=0.0f;  // then don't bother init-ing it
		vertexlist[i].nx=0.0f;
		vertexlist[i].ny=0.0f;
		vertexlist[i].nz=0.0f;
		vertexlist[i].originalvertex=0;
		vertexlist[i].x=0.0f;
		vertexlist[i].y=0.0f;
		vertexlist[i].z=0.0f;
	}

	for (i=0; i < polycount; i++) {

		IEPoly *poly = model->m_polyarray[i];

		// first, figure out the material for this model

		// see if this poly has a texture applied
		vpmap = ValidPolyMap(model,i);

		if (exporttextures && vpmap) {
			// get the map id from the texture_map_id;
			IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

			// get the mat index by sending the mat index and mapnode id
			polylist[i].matlistindex = GetMaterialIndexWithTexture(model, poly->m_attrid, mapnode->m_mapid);
		}
		else	// this poly has no texture, so just set the material to the attrid
			polylist[i].matlistindex = poly->m_attrid;


		
		
		// Cycle through this for each of the poly's 4 vertices
		for (j=0; j<4; j++) {

			if (ingoreNormalVDIs) { // simply overwrite normals so they're not checked below
				// Write the vertex normals no matter what
				vertexlist[poly->m_vertexid[j]].nx=normalarray[poly->m_normalid[j]].x;
				vertexlist[poly->m_vertexid[j]].ny=normalarray[poly->m_normalid[j]].y;
				vertexlist[poly->m_vertexid[j]].nz=normalarray[poly->m_normalid[j]].z;

				// We also need to loop through all duplicate verts to search for any
				// dupe verts that have this vert as it's original vertex, and then
				// overwrite that normal as well
				if (vertcount < totalvertcount) {
					for (int nSearch = vertcount; nSearch < totalvertcount; nSearch++) {
						if (vertexlist[nSearch].originalvertex == poly->m_vertexid[j]) {
							// Write the vertex normals no matter what
							vertexlist[nSearch].nx=normalarray[poly->m_normalid[j]].x;
							vertexlist[nSearch].ny=normalarray[poly->m_normalid[j]].y;
							vertexlist[nSearch].nz=normalarray[poly->m_normalid[j]].z;
						}
					}
				}
			}

			// CASE NUMBER ONE: Vertex not written-just fill in the data
			if (!vertexlist[poly->m_vertexid[j]].modified) {
				// check if vertex has a mapnode
				if ( vpmap ) { // has a mapnode
					// write the color map UVs
					IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];
					// Get the repeat setting
					float rX = mapnode->m_repeat.x;
					float rY = mapnode->m_repeat.y;

					vertexlist[poly->m_vertexid[j]].u=mapnode->m_uv[j].x * rX;
					vertexlist[poly->m_vertexid[j]].v=mapnode->m_uv[j].y * rY;
				}
				else {  // no mapnode, so set them to 0.0f
					vertexlist[poly->m_vertexid[j]].u=0.000000;
					vertexlist[poly->m_vertexid[j]].v=0.000000;
				}

				// Write the vertex normals no matter what
				vertexlist[poly->m_vertexid[j]].nx=normalarray[poly->m_normalid[j]].x;
				vertexlist[poly->m_vertexid[j]].ny=normalarray[poly->m_normalid[j]].y;
				vertexlist[poly->m_vertexid[j]].nz=normalarray[poly->m_normalid[j]].z;

				// Write the vertex coords
				vertexlist[poly->m_vertexid[j]].x=vertexarray[poly->m_vertexid[j]].x;
				vertexlist[poly->m_vertexid[j]].y=vertexarray[poly->m_vertexid[j]].y;
				vertexlist[poly->m_vertexid[j]].z=vertexarray[poly->m_vertexid[j]].z;

				// Write the original vertex
				vertexlist[poly->m_vertexid[j]].originalvertex=poly->m_vertexid[j];

				// Change modified value to TRUE
				vertexlist[poly->m_vertexid[j]].modified=TRUE;

				// Set the vertex entry for the polylist
				polylist[i].vertex[j]=poly->m_vertexid[j];
			}
			else {
				// CASE NUMBER TWO: Written already-check all possibilities
				
				// First, get the uvsexist, uvsmatch and normalsmatch values
				if ( vpmap ) { // UVsExist
					uvsexist=TRUE;
					IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

					// Get the repeat setting
					float rX = mapnode->m_repeat.x;
					float rY = mapnode->m_repeat.y;

					//now check uvsmatch
					if(	(vertexlist[poly->m_vertexid[j]].u==(mapnode->m_uv[j].x * rX)) &&
						(vertexlist[poly->m_vertexid[j]].v==(mapnode->m_uv[j].y * rY)))
						uvsmatch=TRUE;
					else
						uvsmatch=FALSE;
				}
				else
					uvsexist=FALSE;

				if(	(vertexlist[poly->m_vertexid[j]].nx==normalarray[poly->m_normalid[j]].x) &&
					(vertexlist[poly->m_vertexid[j]].ny==normalarray[poly->m_normalid[j]].y) &&
					(vertexlist[poly->m_vertexid[j]].nz==normalarray[poly->m_normalid[j]].z))
					normalsmatch=TRUE;
				else
					normalsmatch=FALSE;

				// Now, cycle through all the possibilities

				//	Case 1) UV exist/match        and  Normals match
				//			-do nothing, it's a match
				if(uvsexist && uvsmatch && normalsmatch) {
					// Set the vertex entry for the polylist
					polylist[i].vertex[j]=poly->m_vertexid[j];
				}

				//	Case 2) UV exist/match        and  Normals don't match
				//			-loop through new verts
				//				-if match , set origvertex number
				//				-if not, create new vertex with UV
				//	Case 3) UV exist/don't match  and  Normals match
				//			-loop through new verts
				//				-if match , set origvertex number
				//				-if not, create new vertex with UV
				//	Case 4) UV exist/don't match  and  Normals don't match
				//			-loop through new verts
				//				-if match , set origvertex number
				//				-if not, create new vertex with UV
				//	Case 5) UV don't exist		and  Normals don't match
				//			-loop through new verts
				//				-if match , set origvertex number
				//				-if not, create a new vertex with UV=0,0
				//	Case 6) UV don't exist		and  Normals Match
				//			-if UV=0,0, do nothing (it's a match)
				//			-if UV!=0,0, loop through new verts
				//				-if match , set origvertex number
				//				-if not, create a new vertex with UV=0,0

				// can this if loop be replaced with just an else
				// because it covers every other possible case?
				// First check the possibility of case 6a
				if (!uvsexist
					&& normalsmatch
					&& FLOAT_EQ(vertexlist[poly->m_vertexid[j]].u,0.000000)
					&& FLOAT_EQ(vertexlist[poly->m_vertexid[j]].v,0.000000)) {
					//it's a match for 6a
					// Set the vertex entry for the polylist
					polylist[i].vertex[j]=poly->m_vertexid[j];
				}
				else {
					// perform the loop/match/create check
					if (vertcount == totalvertcount) {
						// no new verts added, so nothing to check
						// against -- just create a new one
						verttoadd.modified=1;
						if (uvsexist) { // there's a uv to be used
							verttoadd.u=poly->m_mapnodearray[texture_map_id]->m_uv[j].x * poly->m_mapnodearray[texture_map_id]->m_repeat.x;
							verttoadd.v=poly->m_mapnodearray[texture_map_id]->m_uv[j].y * poly->m_mapnodearray[texture_map_id]->m_repeat.y;
						}
						else { // no uvs here, so set them to 0.0f
							verttoadd.u=0.000000;
							verttoadd.v=0.000000;
						}
						verttoadd.nx=normalarray[poly->m_normalid[j]].x;
						verttoadd.ny=normalarray[poly->m_normalid[j]].y;
						verttoadd.nz=normalarray[poly->m_normalid[j]].z;

						// Write the vertex coords
						verttoadd.x=vertexarray[poly->m_vertexid[j]].x;
						verttoadd.y=vertexarray[poly->m_vertexid[j]].y;
						verttoadd.z=vertexarray[poly->m_vertexid[j]].z;

						verttoadd.originalvertex=poly->m_vertexid[j];
						vertexlist.Add(verttoadd);
						polylist[i].vertex[j]=totalvertcount;
						totalvertcount++; // add to the total count
					}
					else { // there are new verts to be checked against

						foundmatch=FALSE;  // reset the foundmatch boolean

						// Loop through the new verts
						for(k=vertcount; k<totalvertcount; k++) { 
							if (!foundmatch &&
								(vertexlist[k].originalvertex == poly->m_vertexid[j]) &&  // make sure it's the same vertex
								(FLOAT_EQ(normalarray[poly->m_normalid[j]].x,vertexlist[k].nx)) &&
								(FLOAT_EQ(normalarray[poly->m_normalid[j]].y,vertexlist[k].ny)) &&
								(FLOAT_EQ(normalarray[poly->m_normalid[j]].z,vertexlist[k].nz)) &&
								(FLOAT_EQ(vertexarray[poly->m_vertexid[j]].x,vertexlist[k].x)) &&
								(FLOAT_EQ(vertexarray[poly->m_vertexid[j]].y,vertexlist[k].y)) &&
								(FLOAT_EQ(vertexarray[poly->m_vertexid[j]].z,vertexlist[k].z))) {
								if (uvsexist) {
									if ((FLOAT_EQ((poly->m_mapnodearray[texture_map_id]->m_uv[j].x * poly->m_mapnodearray[texture_map_id]->m_repeat.x),vertexlist[k].u)) &&
										(FLOAT_EQ((poly->m_mapnodearray[texture_map_id]->m_uv[j].y * poly->m_mapnodearray[texture_map_id]->m_repeat.y),vertexlist[k].v))) {
										//this one is a match
										// Set the vertex entry for the polylist
										polylist[i].vertex[j]=k;
										// set foundmatch to TRUE
										foundmatch=TRUE;
									}
								}
								else { // uvs don't exist, so check against 0
									if ((FLOAT_EQ(0.000000,vertexlist[k].u)) &&
										(FLOAT_EQ(0.000000,vertexlist[k].v))) {
										//this one is a match
										// Set the vertex entry for the polylist
										polylist[i].vertex[j]=k;
										// set foundmatch to TRUE
										foundmatch=TRUE;
									}
								}
							}
						}
						if (!foundmatch) {

							// no match found-make new vert
							verttoadd.modified=1;
							if (uvsexist) { // there's a uv to be used
								verttoadd.u=poly->m_mapnodearray[texture_map_id]->m_uv[j].x * poly->m_mapnodearray[texture_map_id]->m_repeat.x;
								verttoadd.v=poly->m_mapnodearray[texture_map_id]->m_uv[j].y * poly->m_mapnodearray[texture_map_id]->m_repeat.y;
							}
							else { // no uvs here, so set them to 0.0f
								verttoadd.u=0.000000;
								verttoadd.v=0.000000;
							}
							verttoadd.nx=normalarray[poly->m_normalid[j]].x;
							verttoadd.ny=normalarray[poly->m_normalid[j]].y;
							verttoadd.nz=normalarray[poly->m_normalid[j]].z;

							// Write the vertex coords
							verttoadd.x=vertexarray[poly->m_vertexid[j]].x;
							verttoadd.y=vertexarray[poly->m_vertexid[j]].y;
							verttoadd.z=vertexarray[poly->m_vertexid[j]].z;

							verttoadd.originalvertex=poly->m_vertexid[j];
							vertexlist.Add(verttoadd);
							polylist[i].vertex[j]=totalvertcount;
							totalvertcount++; // add to the total count
						}
					}
				}
			}
        }
    }
}

void CHxtApp::GetVertexPolyData(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
	VERTEXDATA verttoadd;

	BOOL uvsexist=FALSE;	// For faster code, these are 
	BOOL uvsmatch=FALSE;	// calculated only once per loop

	BOOL foundmatch = FALSE;  // used in for loop

	int j, k; // loop counters

	// Get the various initial CArray sizes
	int vertcount = model->m_vertexlist.GetSize();
	totalvertcount = vertcount;
	int polycount = model->m_polyarray.GetSize();

	Vector *vertexarray = model->m_vertexlist.GetArray();

	// Set the CArray sizes
	vertexlist.SetSize(vertcount); // This one will likely grow
	polylist.SetSize(polycount);

	vertcount = min(65536, vertcount);

	BOOL vpmap = FALSE;

	// Initialize the vertexlist modified count
	// Only init these values to save time, as the program
	// will never check other fields if modified=0;

	for (int i=0; i < vertcount; i++) {
		vertexlist[i].modified=0;
		vertexlist[i].u=0.0f;  // if all situations write this,
		vertexlist[i].v=0.0f;  // then don't bother init-ing it
		vertexlist[i].originalvertex=0;
		vertexlist[i].x=0.0f;
		vertexlist[i].y=0.0f;
		vertexlist[i].z=0.0f;
	}

	for (i=0; i < polycount; i++) {

		IEPoly *poly = model->m_polyarray[i];

		// first, figure out the material for this model

		// see if this poly has a texture applied
		vpmap = ValidPolyMap(model,i);

		if (exporttextures && vpmap) {
			// get the map id from the texture_map_id;
			IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

			// get the mat index by sending the mat index and mapnode id
			polylist[i].matlistindex = GetMaterialIndexWithTexture(model, poly->m_attrid, mapnode->m_mapid);
		}
		else	// this poly has no texture, so just set the material to the attrid
			polylist[i].matlistindex = poly->m_attrid;


		// Cycle through this for each of the poly's 4 vertices
		for (j=0; j<4; j++) {
			// CASE NUMBER ONE: Vertex not written-just fill in the data
			if (!vertexlist[poly->m_vertexid[j]].modified) {
				// check if vertex has a mapnode
				if ( vpmap ) { // has a mapnode
					// write the color map UVs
					IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

					// Get the repeat setting
					float rX = mapnode->m_repeat.x;
					float rY = mapnode->m_repeat.y;

					vertexlist[poly->m_vertexid[j]].u=mapnode->m_uv[j].x * rX;
					vertexlist[poly->m_vertexid[j]].v=mapnode->m_uv[j].y * rY;
				}
				else {  // no mapnode, so set them to 0.0f
					vertexlist[poly->m_vertexid[j]].u=0.000000;
					vertexlist[poly->m_vertexid[j]].v=0.000000;
				}

				// Write the vertex coords
				vertexlist[poly->m_vertexid[j]].x=vertexarray[poly->m_vertexid[j]].x;
				vertexlist[poly->m_vertexid[j]].y=vertexarray[poly->m_vertexid[j]].y;
				vertexlist[poly->m_vertexid[j]].z=vertexarray[poly->m_vertexid[j]].z;

				// Write the original vertex
				vertexlist[poly->m_vertexid[j]].originalvertex=poly->m_vertexid[j];

				// Change modified value to TRUE
				vertexlist[poly->m_vertexid[j]].modified=TRUE;

				// Set the vertex entry for the polylist
				polylist[i].vertex[j]=poly->m_vertexid[j];
			}
			else {
				// CASE NUMBER TWO: Written already-check all possibilities
				
				// First, get the uvsexist, uvsmatch and normalsmatch values
				if ( vpmap ) { // UVsExist
					uvsexist=TRUE;
					IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

					// Get the repeat setting
					float rX = mapnode->m_repeat.x;
					float rY = mapnode->m_repeat.y;

					//now check uvsmatch
					if(	(vertexlist[poly->m_vertexid[j]].u==(mapnode->m_uv[j].x * rX)) &&
						(vertexlist[poly->m_vertexid[j]].v==(mapnode->m_uv[j].y * rY)))
						uvsmatch=TRUE;
					else
						uvsmatch=FALSE;
				}
				else
					uvsexist=FALSE;

				// Now, cycle through all the possibilities

				//	Case 1) UV exist/match        and  Normals match
				//			-do nothing, it's a match
				if(uvsexist && uvsmatch) {
					// Set the vertex entry for the polylist
					polylist[i].vertex[j]=poly->m_vertexid[j];
				}

				//	Case 2) UV exist/don't match
				//			-loop through new verts
				//				-if match , set origvertex number
				//				-if not, create new vertex with UV
				//	Case 3) UV don't exist
				//			-if UV=0,0, do nothing (it's a match) //3A
				//			-if UV!=0,0, loop through new verts   //3B
				//				-if match , set origvertex number
				//				-if not, create a new vertex with UV=0,0

				// can this if loop be replaced with just an else
				// because it covers every other possible case?
				// First check the possibility of case 3A
				if (!uvsexist
					&& FLOAT_EQ(vertexlist[poly->m_vertexid[j]].u,0.000000)
					&& FLOAT_EQ(vertexlist[poly->m_vertexid[j]].v,0.000000)) {
					//it's a match for 3A
					// Set the vertex entry for the polylist
					polylist[i].vertex[j]=poly->m_vertexid[j];
				}
				else {
					// perform the loop/match/create check
					if (vertcount == totalvertcount) {
						// no new verts added, so nothing to check
						// against -- just create a new one
						verttoadd.modified=1;
						if (uvsexist) { // there's a uv to be used
							verttoadd.u=poly->m_mapnodearray[texture_map_id]->m_uv[j].x * poly->m_mapnodearray[texture_map_id]->m_repeat.x;
							verttoadd.v=poly->m_mapnodearray[texture_map_id]->m_uv[j].y * poly->m_mapnodearray[texture_map_id]->m_repeat.y;
						}
						else { // no uvs here, so set them to 0.0f
							verttoadd.u=0.000000;
							verttoadd.v=0.000000;
						}

						// Write the vertex coords
						verttoadd.x=vertexarray[poly->m_vertexid[j]].x;
						verttoadd.y=vertexarray[poly->m_vertexid[j]].y;
						verttoadd.z=vertexarray[poly->m_vertexid[j]].z;

						verttoadd.originalvertex=poly->m_vertexid[j];
						vertexlist.Add(verttoadd);
						polylist[i].vertex[j]=totalvertcount;

						totalvertcount++; // add to the total count
					}
					else { // there are new verts to be checked against
						foundmatch=FALSE;  // reset the foundmatch boolean

						// Loop through the new verts
						for(k=vertcount; k<totalvertcount; k++) { 
							if (!foundmatch &&
								(vertexlist[k].originalvertex == poly->m_vertexid[j]) &&  // make sure it's the same vertex
								(FLOAT_EQ(vertexarray[poly->m_vertexid[j]].x,vertexlist[k].x)) &&
								(FLOAT_EQ(vertexarray[poly->m_vertexid[j]].y,vertexlist[k].y)) &&
								(FLOAT_EQ(vertexarray[poly->m_vertexid[j]].z,vertexlist[k].z))) {
								if (uvsexist) {
									if ((FLOAT_EQ((poly->m_mapnodearray[texture_map_id]->m_uv[j].x * poly->m_mapnodearray[texture_map_id]->m_repeat.x),vertexlist[k].u)) &&
										(FLOAT_EQ((poly->m_mapnodearray[texture_map_id]->m_uv[j].y * poly->m_mapnodearray[texture_map_id]->m_repeat.y),vertexlist[k].v))) {
										//this one is a match
										// Set the vertex entry for the polylist
										polylist[i].vertex[j]=k;
										// set foundmatch to TRUE
										foundmatch=TRUE;
									}
								}
								else { // uvs don't exist, so check against 0
									if ((FLOAT_EQ(0.000000,vertexlist[k].u)) &&
										(FLOAT_EQ(0.000000,vertexlist[k].v))) {
										//this one is a match
										// Set the vertex entry for the polylist
										polylist[i].vertex[j]=k;
										// set foundmatch to TRUE
										foundmatch=TRUE;
									}
								}
							}
						}
						if (!foundmatch) {
							// no match found-make new vert
							verttoadd.modified=1;
							if (uvsexist) { // there's a uv to be used
								verttoadd.u=poly->m_mapnodearray[texture_map_id]->m_uv[j].x * poly->m_mapnodearray[texture_map_id]->m_repeat.x;
								verttoadd.v=poly->m_mapnodearray[texture_map_id]->m_uv[j].y * poly->m_mapnodearray[texture_map_id]->m_repeat.y;
							}
							else { // no uvs here, so set them to 0.0f
								verttoadd.u=0.000000;
								verttoadd.v=0.000000;
							}

							// Write the vertex coords
							verttoadd.x=vertexarray[poly->m_vertexid[j]].x;
							verttoadd.y=vertexarray[poly->m_vertexid[j]].y;
							verttoadd.z=vertexarray[poly->m_vertexid[j]].z;

							verttoadd.originalvertex=poly->m_vertexid[j];
							vertexlist.Add(verttoadd);
							polylist[i].vertex[j]=totalvertcount;

							totalvertcount++; // add to the total count
						}
					}
				}
			}
        }
    }
}

void CHxtApp::GetVertexPolyNormalDataNoVDIs(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
	int j; // loop counter

	// Get the various initial CArray sizes
	int vertcount = model->m_vertexlist.GetSize();
	totalvertcount = vertcount;
	int polycount = model->m_polyarray.GetSize();

	Vector *normalarray = model->m_normallist.GetArray();
	Vector *vertexarray = model->m_vertexlist.GetArray();

	// Set the CArray sizes
	vertexlist.SetSize(vertcount); // This one will likely grow
	polylist.SetSize(polycount);

	vertcount = min(65536, vertcount);

	BOOL vpmap = FALSE;

	// Initialize the vertexlist modified count
	// Only init these values to save time, as the program
	// will never check other fields if modified=0;

	for (int i=0; i < vertcount; i++) {
		vertexlist[i].modified=0;
		vertexlist[i].u=0.0f;  // if all situations write this,
		vertexlist[i].v=0.0f;  // then don't bother init-ing it
		vertexlist[i].nx=0.0f;
		vertexlist[i].ny=0.0f;
		vertexlist[i].nz=0.0f;
		vertexlist[i].originalvertex=0;
		vertexlist[i].x=0.0f;
		vertexlist[i].y=0.0f;
		vertexlist[i].z=0.0f;
	}

	for (i=0; i < polycount; i++) {
		IEPoly *poly = model->m_polyarray[i];

		// first, figure out the material for this model

		// see if this poly has a texture applied
		vpmap = ValidPolyMap(model,i);

		if (exporttextures && vpmap) {
			// get the map id from the texture_map_id;
			IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

			// get the mat index by sending the mat index and mapnode id
			polylist[i].matlistindex = GetMaterialIndexWithTexture(model, poly->m_attrid, mapnode->m_mapid);
		}
		else	// this poly has no texture, so just set the material to the attrid
			polylist[i].matlistindex = poly->m_attrid;

		// Cycle through this for each of the poly's 4 vertices
		for (j=0; j<4; j++) {
			if ( vpmap ) { // has a mapnode
				// write the color map UVs
				IEMapNode *mapnode = poly->m_mapnodearray[texture_map_id];

				// Get the repeat setting
				float rX = mapnode->m_repeat.x;
				float rY = mapnode->m_repeat.y;

				vertexlist[poly->m_vertexid[j]].u=mapnode->m_uv[j].x * rX;
				vertexlist[poly->m_vertexid[j]].v=mapnode->m_uv[j].y * rY;
			}
			else {  // no mapnode, so set them to 0.0f
				vertexlist[poly->m_vertexid[j]].u=0.000000;
				vertexlist[poly->m_vertexid[j]].v=0.000000;
			}

			// Write the vertex normals no matter what
			vertexlist[poly->m_vertexid[j]].nx=normalarray[poly->m_normalid[j]].x;
			vertexlist[poly->m_vertexid[j]].ny=normalarray[poly->m_normalid[j]].y;
			vertexlist[poly->m_vertexid[j]].nz=normalarray[poly->m_normalid[j]].z;

			// Write the vertex coords
			vertexlist[poly->m_vertexid[j]].x=vertexarray[poly->m_vertexid[j]].x;
			vertexlist[poly->m_vertexid[j]].y=vertexarray[poly->m_vertexid[j]].y;
			vertexlist[poly->m_vertexid[j]].z=vertexarray[poly->m_vertexid[j]].z;

			// Set the vertex entry for the polylist
			polylist[i].vertex[j]=poly->m_vertexid[j];
        }
    }
}

void CHxtApp::WriteVDIs(FileStream &fs, IEPolyModel *model, LONG &totalitems, LONG &curritem)
{
	int count = model->m_vertexlist.GetSize();

	fs << "VertexDuplicationIndices {" << nl;
	fs << "  " << totalvertcount << ";          // Total Vertex Count" << nl;
	fs << "  " << count << ";          // Original Vertex Count" << nl;

	for (int i=0; i<totalvertcount; i++) {
		fs << "  " << vertexlist[i].originalvertex;

		if (i+1 < totalvertcount)  // If this is NOT the last VDI,
			fs << "," << nl;  // add a coma
	}

	fs << ";" << nl << "}" << nl; // Add the VDI-closing semicolon and bracket
}


void CHxtApp::WriteVDIHeader(FileStream &fs)
{
	fs << "template VertexDuplicationIndices {" << nl;
	fs << " <b8d65549-d7c9-4995-89cf-53a9a8b031e3>" << nl;
	fs << " DWORD nIndices;" << nl;
	fs << " DWORD nOriginalVertices;" << nl;
	fs << " array DWORD indices[nIndices];" << nl;
	fs << "}" << nl << nl;
}

void CHxtApp::WriteBoneHeader(FileStream &fs)
{
	fs << "template XSkinMeshHeader {" << nl;
	fs << " <3cf169ce-ff7c-44ab-93c0-f78f62d172e2>" << nl;
	fs << " WORD nMaxSkinWeightsPerVertex;" << nl;
	fs << " WORD nMaxSkinWeightsPerFace;" << nl;
	fs << " WORD nBones;" << nl;
	fs << "}" << nl << nl;

	fs << "template SkinWeights {" << nl;
	fs << " <6f0d123b-bad2-4167-a0d0-80224f25fabb>" << nl;
	fs << " STRING transformNodeName;" << nl;
	fs << " DWORD nWeights;" << nl;
	fs << " array DWORD vertexIndices[nWeights];" << nl;
	fs << " array float weights[nWeights];" << nl;
	fs << " Matrix4x4 matrixOffset;" << nl;
	fs << "}" << nl << nl;
}

void CHxtApp::WriteFVFDataHeader(FileStream &fs)
{
	fs << "template FVFData {" << nl;
	fs << " <b6e70a0e-8ef9-4e83-94ad-ecc8b0c04897>" << nl;
	fs << " DWORD dwFVF;" << nl;
	fs << " DWORD nDWords;" << nl;
	fs << " array DWORD data[nDWords];" << nl;
	fs << "}" << nl << nl;
}

int IsRMFrameMember(int vertexnum, LONG *array, int arraysize)
{
	for (int i=0;i<arraysize;i++)
	{
		if (vertexnum==array[i])
			return i;
	}

	// no match, so return -1
	return -1;
}

Vector CHxtApp::RotateTransVertexByMatrix(Vector vertex, Matrix44 matrix, Matrix44 absolutematrix)
{
	Vector output;
	Vector4 matrix0, matrix1, matrix2, matrix3;
	Vector4 translation;

	translation = absolutematrix[3];

	matrix0 = matrix[0];
	matrix1 = matrix[1];
	matrix2 = matrix[2];
	matrix3 = matrix[3];

	vertex[0] = vertex[0]-translation[0];
	vertex[1] = vertex[1]-translation[1];
	vertex[2] = vertex[2]-translation[2];

	// Rotate the point by the matrix rotation
	output[0] = ((vertex[0] * matrix0[0]) +
					(vertex[1] * matrix1[0]) +
					(vertex[2] * matrix2[0]) +
					(1 * matrix3[0]));
	output[1] = ((vertex[0] * matrix0[1]) +
					(vertex[1] * matrix1[1]) +
					(vertex[2] * matrix2[1]) +
					(1 * matrix3[1]));
	output[2] = ((vertex[0] * matrix0[2]) +
					(vertex[1] * matrix1[2]) +
					(vertex[2] * matrix2[2]) +
					(1 * matrix3[2]));

	vertex[0] = vertex[0]+translation[0];
	vertex[1] = vertex[1]+translation[1];
	vertex[2] = vertex[2]+translation[2];

	// now translate it by the matrix rotation

//	vertex[0] = vertex[0]-translation[0];
//	vertex[1] = vertex[1]-translation[1];
//	vertex[2] = vertex[2]-translation[2];

	return output;
}

Vector CHxtApp::RotateVectorByMatrix(Vector vector, Matrix44 matrix)
{
	Vector output;
	Vector4 matrix0, matrix1, matrix2, matrix3;

	matrix0 = matrix[0];
	matrix1 = matrix[1];
	matrix2 = matrix[2];
	matrix3 = Vector4(0,0,0,1);

	// Rotate the point by the matrix rotation
	output[0] = ((vector[0] * matrix0[0]) +
					(vector[1] * matrix1[0]) +
					(vector[2] * matrix2[0]) +
					(1 * matrix3[0]));
	output[1] = ((vector[0] * matrix0[1]) +
					(vector[1] * matrix1[1]) +
					(vector[2] * matrix2[1]) +
					(1 * matrix3[1]));
	output[2] = ((vector[0] * matrix0[2]) +
					(vector[1] * matrix1[2]) +
					(vector[2] * matrix2[2]) +
					(1 * matrix3[2]));

	return output;
}

void CHxtApp::SaveRetainedModeBoneBranch(FileStream &fs, IEPolyModel *model, IEBone *bone, String &spaces, Matrix44 Parent, Matrix44 CombinedParent)
{
	Matrix44 BoneTransform;
	Matrix44 NewCombinedTransform;
	Vector4 v;
	char vstring[80];
	Vector4 v1, v2, v3, v4;
	Matrix44 final, final2;
	Matrix44 xrot, yrot, zrot;
	int x;
	int j;
	int k;
	DUPEVERT newdupevert;
	
	Vector newvertex, currentvertex;

	int totalboneverts;

	RMPOLYMATINFO newrmpolymat;

	int found;

	Matrix44 AbsoluteBoneMat;

	char floatstring[80];

	int RMnewmatcount = 0;

	int texmatnum = 0;

	int orig_attr_count = model->m_attrlist.GetSize();

	IEMap *maparray = model->m_maplist.GetArray();
	char texturename[128];

	// transposed rotations
	v1.Set(1.000000,0.000000,0.000000,0.000000);
	v2.Set(0.000000,0.000000,-1.000000,0.000000);
	v3.Set(0.000000,1.000000,0.000000,0.000000);
	v4.Set(0.000000,0.000000,0.000000,1.000000);
	xrot = Matrix44(v1, v2, v3, v4);

	v1.Set(0.000000,0.000000,1.000000,0.000000);
	v2.Set(0.000000,1.000000,0.000000,0.000000);
	v3.Set(-1.000000,0.000000,0.000000,0.000000);
	v4.Set(0.000000,0.000000,0.000000,1.000000);
	yrot = Matrix44(v1, v2, v3, v4);

	v1.Set(0.000000,-1.000000,0.000000,0.000000);
	v2.Set(1.000000,0.000000,0.000000,0.000000);
	v3.Set(0.000000,0.000000,1.000000,0.000000);
	v4.Set(0.000000,0.000000,0.000000,1.000000);
	zrot = Matrix44(v1, v2, v3, v4);

	for (; bone; bone = bone->m_sibling) {

	  ComputeBoneToStandard(bone->m_pivotpos, bone->m_endpos, bone->m_rollpos);

	  AbsoluteBoneMat = ToMatrix44 ( BoneMatrix );

	  BoneTransform = GetMatrixConversion(Parent, AbsoluteBoneMat);
	  BoneTransform = BoneTransform.Transpose();
	  BoneTransform = BoneTransform.Inverse();  // new 10/29 from import/export help file

	  // Populate the Bonelist skinweight values
	  // first get a valid directx name
	  CString tempBoneName = bone->m_name;
	  tempBoneName = GetValidDirectXName(tempBoneName);
	  strcpy(bonelist[bonenumber].name, tempBoneName);

	  NewCombinedTransform =  BoneTransform * CombinedParent;

	  bonelist[bonenumber].SkinWeightOffset = NewCombinedTransform.Inverse();

	  // Output the frame info
  	  fs << nl << spaces << "Frame ";

	  CString BoneName = bone->m_name;
	  BoneName = GetValidDirectXName(BoneName);
	  fs.Write( BoneName, BoneName.GetLength());

	  fs << " {" << nl;

  	  fs << spaces << "  FrameTransformMatrix {" << nl;

	  for (x=0;x<4;x++) {
		  v = BoneTransform[x];
		  sprintf(vstring, "    %.6f,%.6f,%.6f,%.6f", v[0],v[1],v[2],v[3]);
	  	  fs << spaces << vstring;
		  if (x < 3)
			  fs << "," << nl;
		  else
			  fs << ";;" << nl << spaces << "  }" << nl;
	  }

	  int boneidcount = bone->m_vertexidarray.GetSize();

	  if (boneidcount == 0)  // no verts in this bone
		  //Increase the bonenumber for use in the bonelist
		  bonenumber++;
	  else { // handle the frame
		  bonenumber++;

		  LONG *array = bone->m_vertexidarray.GetArray();

		  if (exportVDIs) {

			  // cycle through each VDI
			  for (int i=initialvertcount;i<totalvertcount;i++) {
				  // for each vdi, cycle through the bone's vertex list for matches
				  for (j=0;j<boneidcount;j++) {
					  if (vertexlist[i].originalvertex == array[j]) {
						  newdupevert.vertindex = i;
					  	  newdupevert.bonevertindex = j;
						  dupeverts.Add(newdupevert);
					  }
				  }
			  }

	  		  totalboneverts = boneidcount + dupeverts.GetSize();
		  }
		  else
			  totalboneverts = boneidcount;


		  // output the mesh info for this bone
	  	  fs << nl << spaces << "  " << "Mesh {" << nl;
		  fs << spaces << "  " << totalboneverts << ";" << nl;

		  // Output the original vertex id list
		  for (int i=0; i < boneidcount; i++) {

			  currentvertex[0] = vertexlist[array[i]].x;
			  currentvertex[1] = vertexlist[array[i]].y;
			  currentvertex[2] = vertexlist[array[i]].z;// * modelorientation;

			  // convert the point using the bones trans/rot matrix
			  newvertex = RotateTransVertexByMatrix(currentvertex, NewCombinedTransform.Inverse(), AbsoluteBoneMat);

			  sprintf(floatstring, "%.6f;%.6f;%.6f;",
				  newvertex[0],
				  newvertex[1],
				  newvertex[2] );
			  fs  << spaces << "  " << floatstring;

			  if (i < (boneidcount-1))
				  fs << "," << nl;
		  }

		  if ((!exportVDIs) ||(exportVDIs && (boneidcount == totalboneverts)))
			  // no vdis, so close the vertexlist
			  fs << ";" << nl;
		  else {
			  fs << "," << nl; // close out the previous non-vdi vertex

			  // output the VDI verts
			  for (i=0; i < dupeverts.GetSize();i++) {
	  			  currentvertex[0] = vertexlist[dupeverts[i].vertindex].x;
				  currentvertex[1] = vertexlist[dupeverts[i].vertindex].y;
				  currentvertex[2] = vertexlist[dupeverts[i].vertindex].z;// * modelorientation;

				  // convert the point using the bones trans/rot matrix
				  newvertex = RotateTransVertexByMatrix(currentvertex, NewCombinedTransform.Inverse(), AbsoluteBoneMat);

				  sprintf(floatstring, "%.6f;%.6f;%.6f;",
				  newvertex[0],
				  newvertex[1],
				  newvertex[2] );

				  fs  << spaces << "  " << floatstring;

				  if (i < (dupeverts.GetSize()-1))
					  fs << "," << nl;
				  else
					  fs << ";" << nl;
			  }
		  }

		  // figure out which polys are made from this vertex list and display them
		  // cycle through poly list
		  for (i=0;i<polylist.GetSize();i++) {
			  int v1,v2,v3,v4;

			  // this function returns the vertex's position inside this bone
			  v1=IsRMFrameMember(polylist[i].vertex[0], bone->m_vertexidarray.GetArray(), boneidcount);
			  v2=IsRMFrameMember(polylist[i].vertex[1], bone->m_vertexidarray.GetArray(), boneidcount);
			  v3=IsRMFrameMember(polylist[i].vertex[2], bone->m_vertexidarray.GetArray(), boneidcount);
			  v4=IsRMFrameMember(polylist[i].vertex[3], bone->m_vertexidarray.GetArray(), boneidcount);

			  // must also check the vdi list
			  if (v1 == -1)
				  for (j=0;j<dupeverts.GetSize();j++)
					  if (polylist[i].vertex[0] == dupeverts[j].vertindex)
						  v1 = boneidcount + j; // it's the vertex j places past the bonevertlist

			  if (v2 == -1)
				  for (j=0;j<dupeverts.GetSize();j++)
					  if (polylist[i].vertex[1] == dupeverts[j].vertindex)
						  v2 = boneidcount + j; // it's the vertex j places past the bonevertlist

			  if (v3 == -1)
				  for (j=0;j<dupeverts.GetSize();j++)
					  if (polylist[i].vertex[2] == dupeverts[j].vertindex)
						  v3 = boneidcount + j; // it's the vertex j places past the bonevertlist

			  if (v4 == -1)
				  for (j=0;j<dupeverts.GetSize();j++)
					  if (polylist[i].vertex[3] == dupeverts[j].vertindex)
						  v4 = boneidcount + j; // it's the vertex j places past the bonevertlist

			  // if all four poly vertices are inside the bonearray
			  if ((v1 > -1) && (v2 > -1) && (v3 > -1) && (v4 > -1)) {
				  newrmpolymat.poly = i;
				  retainedpolylist.Add(newrmpolymat);
				  polylist[i].RMvertex[0]=v1;
				  polylist[i].RMvertex[1]=v2;
				  polylist[i].RMvertex[2]=v3;
				  polylist[i].RMvertex[3]=v4;
			  }
		  }

		  fs << spaces << "  " << retainedpolylist.GetSize() << ";  // number of polys" << nl;
	  
		  for (i=0;i<retainedpolylist.GetSize();i++)
		  {
			  j = retainedpolylist[i].poly;


			  if (FLOAT_EQ(modelorientation, 1.0)) // if it's right handed
				  fs  << spaces << "  " 
					<< "4;"
			  		<< polylist[j].RMvertex[0] << ","
			  		<< polylist[j].RMvertex[1] << ","
					<< polylist[j].RMvertex[2] << ","
					<< polylist[j].RMvertex[3] << ";";
			  else
				  fs  << spaces << "  " 
					<< "4;"
			  		<< polylist[j].RMvertex[3] << ","
			  		<< polylist[j].RMvertex[2] << ","
					<< polylist[j].RMvertex[1] << ","
					<< polylist[j].RMvertex[0] << ";";

			  if (i < (retainedpolylist.GetSize()-1))
				  fs << "," << nl;
			  else
				  fs << ";" << nl;
		  }

		  // get the normals for the vertices in this list

		  if (exportnormals) {
			  fs << nl << spaces << "  " << "MeshNormals {" << nl;
			  fs << spaces << "  " << totalboneverts << ";" << nl;

			  for (i=0; i < boneidcount; i++) {
				  currentvertex[0] = vertexlist[array[i]].nx;
				  currentvertex[1] = vertexlist[array[i]].ny;
				  currentvertex[2] = vertexlist[array[i]].nz;// * modelorientation;

				  // convert the normal using the bones trans/rot matrix
				  newvertex = RotateVectorByMatrix(currentvertex, NewCombinedTransform.Inverse());

				  newvertex = newvertex.Normalize();

				  sprintf(floatstring, "%.6f;%.6f;%.6f;",
					  newvertex[0],
					  newvertex[1],
					  newvertex[2] );
				  fs  << spaces << "  " << floatstring;
				  
				  if (i < (boneidcount-1))
					  fs << "," << nl;
			  }

			  // if no extra verts, close out the normal list
			  if ((!exportVDIs) ||(exportVDIs && (boneidcount == totalboneverts)))
				  fs << ";" << nl;
			  else { // there are vdis, so output them as well
				  fs << "," << nl; // close out the previous non-vdi vertex
				  
				  for (i=0; i < dupeverts.GetSize();i++) {
					  currentvertex[0] = vertexlist[dupeverts[i].vertindex].nx;
					  currentvertex[1] = vertexlist[dupeverts[i].vertindex].ny;
					  currentvertex[2] = vertexlist[dupeverts[i].vertindex].nz;// * modelorientation;

					  // convert the point using the bones trans/rot matrix
					  newvertex = RotateVectorByMatrix(currentvertex, NewCombinedTransform.Inverse());

					  newvertex = newvertex.Normalize();

					  sprintf(floatstring, "%.6f;%.6f;%.6f;",
						  newvertex[0],
						  newvertex[1],
						  newvertex[2] );

					  fs  << spaces << "  " << floatstring;

					  if (i < (dupeverts.GetSize()-1))
						  fs << "," << nl;
					  else
						  fs << ";" << nl;
				  }
			  }

			  // output the normals for the polys in this vertex list
			  fs << spaces << "  " << retainedpolylist.GetSize() << ";  // number of polys" << nl;

			  for (i=0;i<retainedpolylist.GetSize();i++) {
				  j = retainedpolylist[i].poly;


				  if (FLOAT_EQ(modelorientation, 1.0)) // if it's right handed
				  
					  fs  << spaces << "  " 
						<< "4;"
			  			<< polylist[j].RMvertex[0] << ","
			  			<< polylist[j].RMvertex[1] << ","
						<< polylist[j].RMvertex[2] << ","
						<< polylist[j].RMvertex[3] << ";";
				  else
					  fs  << spaces << "  " 
						<< "4;"
			  			<< polylist[j].RMvertex[3] << ","
			  			<< polylist[j].RMvertex[2] << ","
						<< polylist[j].RMvertex[1] << ","
						<< polylist[j].RMvertex[0] << ";";

				  if (i < (retainedpolylist.GetSize()-1))
					  fs << "," << nl;
				  else
					  fs << ";" << nl;
			  }

			  // close the normal list
			  fs << spaces << "  " << "} // end Normallist" << nl << nl;
		  }

		  // get the list of materials that these polys use

		  // cycle through the polys, getting their attr ids
		  // (already optimized from skinned mesh generation)

		  for (i=0;i<retainedpolylist.GetSize();i++)
		  {
  			  j = retainedpolylist[i].poly;
			  found = -1;

			  for (k=0;k<retainedmatlist.GetSize();k++) {
				  if (polylist[j].matlistindex == retainedmatlist[k])
					  found = k;
			  }

			  if (found == -1) {
				  // add the entry to the mat list
				  int newid = polylist[j].matlistindex;
				  retainedmatlist.Add(newid);
				  retainedpolylist[i].matnum = retainedmatlist.GetSize() - 1;
			  }
			  else
				  retainedpolylist[i].matnum = found;
			  
		  }

	 	  // output the list and descriptions of the materials
		  fs << nl << spaces << "  " << "MeshMaterialList {              //Material List" << nl;
		  fs << spaces << "  " << retainedmatlist.GetSize() << ";" << nl;
		  fs << spaces << "  " << retainedpolylist.GetSize() << ";" << nl;

		  for (i=0; i < retainedpolylist.GetSize(); i++) {
			  // output the material number
			  fs << spaces << "  " << retainedpolylist[i].matnum;

			  if (i+1 < retainedpolylist.GetSize())         // if this is NOT the last vertex point,
				  fs << "," << nl;  // add a coma
		  }

		  fs << ";" << nl;  // Add the vertex-closing semicolon

		  // Output the materials   

		  for (i=0; i < retainedmatlist.GetSize(); i++) {
			  j = retainedmatlist[i];

			  fs << nl << spaces << "Material {          // Material" << nl;
			  sprintf(floatstring, "%.6f;%.6f;%.6f;%.6f;;",
				  matlist[j].red,
				  matlist[j].green,
				  matlist[j].blue,
				  matlist[j].alpha);

			  fs  << spaces << floatstring << nl;

			  if (matlist[j].texture == -1) {
				  if(FLOAT_EQ(0.000000,matlist[j].specular)) {
					  fs << spaces << "0.000000;" << nl;
					  fs << spaces << "0.000000;0.000000;0.000000;;" << nl;
				  }

				  else {
					  fs << spaces << "10.000000;" << nl;
					  sprintf(floatstring, "%.6f;%.6f;%.6f;;",
						  matlist[j].specular,
						  matlist[j].specular,
						  matlist[j].specular);
					  
					  fs << spaces << floatstring << nl;

				  }

				  fs << spaces << "0.000000;0.000000;0.000000;;" << nl;
			  
				  fs << spaces << "} // End Material" << nl;
			  }
			  else { // it's a textured material
				  fs << spaces << "299.000000;" << nl;

				  fs << spaces << "0.000000;0.000000;0.000000;;" << nl;
				  fs << spaces << "0.000000;0.000000;0.000000;;" << nl;

				  strcpy(texturename, RemovePathname(fs, maparray[matlist[j].texture].m_filename));
				  fs << spaces << "TextureFilename {\"" << texturename << "\";}" << nl;
				  fs << spaces << "} // End Material" << nl;
			  }
		  }

		  fs << spaces << "} // End MeshMaterialList" << nl;
		  
		  // output the texture coordinates for each vertex in this list

		  if (exporttextures) {
			  fs << nl << spaces << "  " << "MeshTextureCoords {" << nl;
			  fs << spaces << "  " << totalboneverts << ";" << nl;

			  for (i=0; i < boneidcount; i++) {
				  sprintf(floatstring, "%.6f;%.6f;",
					  vertexlist[array[i]].u,
					  vertexlist[array[i]].v );
				  fs  << spaces << "  " << floatstring;

				  if (i < (boneidcount-1))
					  fs << "," << nl;
			  }

			  if ((!exportVDIs) ||(exportVDIs && (boneidcount == totalboneverts)))
				  fs << ";" << nl;
			  else {
				  fs << "," << nl; // close out the previous non-vdi vertex

				  for (i=0; i < dupeverts.GetSize();i++) {
					  sprintf(floatstring, "%.6f;%.6f;",
						  vertexlist[dupeverts[i].vertindex].u,
						  vertexlist[dupeverts[i].vertindex].v );
					  fs  << spaces << "  " << floatstring;

					  if (i < (dupeverts.GetSize()-1))
						  fs << "," << nl;
					  else
						  fs << ";" << nl;
				  }
			  }

			  fs << spaces << "  " << "} // end MeshTextureCoords" << nl << nl;
		  }

		  // export the VDI list if neccessary
		  if ((exportVDIs) && (totalboneverts > boneidcount)) {
			  fs << spaces << "  " << "VertexDuplicationIndices {" << nl;
			  fs << spaces << "  " << "  " << totalboneverts << ";          // Total Vertex Count" << nl;
			  fs << spaces << "  " << "  " << boneidcount << ";          // Original Vertex Count" << nl;

			  for (i=0; i<boneidcount; i++) {
				  fs << spaces << "  " << "  " << i;

				  fs << "," << nl;  // add a coma
			  }

			  // now, output the VDI list
			  for (i=0; i<dupeverts.GetSize();i++) {
				  fs << spaces << "  " << "  " << dupeverts[i].bonevertindex;
				  if (i+1 < dupeverts.GetSize())
					  fs << "," << nl;
			  }

			  fs << ";" << nl << spaces << "  " << "} // End VertexDuplicationIndices" << nl; // Add the VDI-closing semicolon and bracket

		  }

		  fs << spaces << "  " << "} // End mesh" << nl;
   
	  }

	  // clear the lists before continuing
	  retainedpolylist.SetSize(0);
	  retainedpolylist.RemoveAll();
	  retainedpolylist.FreeExtra();
	  retainedmatlist.SetSize(0);
	  retainedmatlist.RemoveAll();
	  retainedmatlist.FreeExtra();
	  dupeverts.SetSize(0);
	  dupeverts.RemoveAll();
	  dupeverts.FreeExtra();
	  RMnewmats.SetSize(0);
	  RMnewmats.RemoveAll();
	  RMnewmats.FreeExtra();

	  if (bone->m_child) {
		  String childspaces = spaces + "  ";
		  SaveRetainedModeBoneBranch( fs, model, bone->m_child, childspaces, AbsoluteBoneMat, NewCombinedTransform );
	  }

	  if (spaces.length() > 0) {		 // if it's not the root frame,
		  fs << nl << spaces << "} // End ";;
		  CString BoneName = bone->m_name;
		  BoneName = GetValidDirectXName(BoneName);

		  fs.Write( BoneName, BoneName.GetLength());
		  fs << " Frame" << nl; // close the frame
	  }
   }
}

void CHxtApp::SaveRetainedModeBones(FileStream &fs, IEPolyModel *model)
{
	Vector4 v1,v2, v3, v4;

	LONG bc = CountBones(model);

	bonelist.SetSize(bc); // list of bones and their SkinWeightOffsets


	Matrix44 offset;
	String spaces = "";

	offset.SetIdentity(); // set initial bone to identity matrix

	if (modelorientation < 0) {
		v1 = offset[0];
		v2 = offset[1];
		v3 = offset[2];
		v4 = offset[3];

		v3[2] = -1.0f;
		
		offset = Matrix44(v1, v2, v3, v4);
	}

	// only use offset in both places for testing purposes
	// change it to another id matrix later on,
	// (when root offset isn't neccesarily the id matrix)
	SaveRetainedModeBoneBranch(fs, model, model->m_bonehead, spaces, offset, offset);

	// Now that we've stored all the bones, check if duplicate bone names
	// have been created due to the valid DirectX name conversion
	DuplicateNameCheck();

}


BOOL CHxtApp::ExportRetainedModeModel(IEPolyModel *model, const String &filename)
{
   FileStream fs; //cjr
   LONG totalitems, curritem = 0;

   newmatcount = 0;
   unusedmats = 0;
   texture_map_id = 0;
   totalvertcount = 0;
   bonenumber = 0;

   polynum = new POLYINFO[model->m_polyarray.GetSize()];

   int newpolynum=model->m_polyarray.GetSize();
   
   InitNewPolys(newpolynum);

   InitMatList(model);

   m_hprogressbar = GetProgressBar();
   if (m_hprogressbar) {
      m_hprogressbar->Initialize();
      CString text;
      text.LoadString(IDS_EXPORTTITLE);
      m_hprogressbar->SetMessage1((LPCTSTR)text);
      m_hprogressbar->SetMessage2((LPCTSTR)filename.GetNameOnly().Get());
   }

   // Get total items for percent bar
   totalitems = model->m_vertexlist.GetSize();

   initialvertcount = model->m_vertexlist.GetSize();

   if (!fs.Open( (String)filename, HashStream::HS_STORE))
      return FALSE;

   if(exportnormals && exportVDIs)
	   GetVertexPolyNormalData(fs, model, totalitems, curritem);
   else if (!exportnormals && exportVDIs)
	   GetVertexPolyData(fs, model, totalitems, curritem);
   else if (!exportVDIs)
	   GetVertexPolyNormalDataNoVDIs(fs, model, totalitems, curritem);

   fs << "xof 0303txt 0032" << nl << nl;

   //If needed, write the VDI template to the file
   if(totalvertcount > model->m_vertexlist.GetSize())
	   WriteVDIHeader(fs);

   WriteFVFDataHeader(fs);

   // Calculate the texture stuff

   if (exportmaterials && (model->m_attrlist.GetSize() > 0))
	   MaterialCleanup(fs, model);

   SaveRetainedModeBones(fs, model);

   fs << "} // End RootFrame" << nl;

   delete [] polynum;
   delete [] matcount;

   if (m_hprogressbar) {
      m_hprogressbar->Uninitialize();
      m_hprogressbar = NULL;
   }


   // Clear the vertexlist and polylist CArrays

   vertexlist.SetSize(0);
   vertexlist.RemoveAll();
   vertexlist.FreeExtra();
   polylist.SetSize(0);
   polylist.RemoveAll();
   polylist.FreeExtra();
   bonelist.SetSize(0);
   bonelist.RemoveAll();
   bonelist.FreeExtra();
   matlist.SetSize(0);
   matlist.RemoveAll();
   matlist.FreeExtra();

   return TRUE;
}

void CHxtApp::InitMatList(IEPolyModel *model)
{
	int nummats = model->m_attrlist.GetSize();
	Attr *attrarray = model->m_attrlist.GetArray();

	matlist.SetSize(nummats);

	for (int i=0;i<nummats;i++) {
		matlist[i].red = attrarray[i].m_diffusecolor.m_red;
		matlist[i].green = attrarray[i].m_diffusecolor.m_green;
		matlist[i].blue = attrarray[i].m_diffusecolor.m_blue;
		matlist[i].alpha = 1.0f - attrarray[i].m_transparency;
		matlist[i].specular = attrarray[i].m_specularintensity;
		matlist[i].texture = -1; // has no texture yet
	}
}

int CHxtApp::GetMaterialIndexWithTexture(IEPolyModel *model, int attrid, int textureid)
{
	// this function searches for an existing texture map material
	// and returns the matlist index
	// if it doesn't find a match, it creates a new one and returns the
	// new index as the matlist index
	Attr *attrarray = model->m_attrlist.GetArray();
	int nummats = model->m_attrlist.GetSize();

	IEMap *maparray = model->m_maplist.GetArray();

	MATERIALINFO mat;

	MATERIALINFO mat2;

	if (texturefacewhite) {
		mat.red		= 1.000000;
		mat.green	= 1.000000;
		mat.blue	= 1.000000;
	}
	else {
		mat.red		= attrarray[attrid].m_diffusecolor.m_red;
		mat.green	= attrarray[attrid].m_diffusecolor.m_green;
		mat.blue	= attrarray[attrid].m_diffusecolor.m_blue;
	}

	mat.alpha	= 1.000000;
	mat.specular = 299.000000;
	mat.texture	= textureid; // has no texture yet

	int nummmm = matlist.GetSize();

	for (int i=nummats;i<matlist.GetSize();i++) {
		// start the loop at nummats, because this function is only called when
		// a poly has a texture applied (skipping past nummats skips the non-textured
		// materials)

		mat2 = matlist[i];
		
		BOOL match = TRUE;

		if (!FLOAT_EQ(mat.red, matlist[i].red))
			match = FALSE;

		if (!FLOAT_EQ(mat.green, matlist[i].green))
			match = FALSE;

		if (!FLOAT_EQ(mat.blue, matlist[i].blue))
			match = FALSE;

		if (!FLOAT_EQ(mat.alpha, matlist[i].alpha))
			match = FALSE;

		if (!FLOAT_EQ(mat.specular, matlist[i].specular))
			match = FALSE;

		CString map1 = maparray[mat.texture].m_filename;
		CString map2 = maparray[matlist[i].texture].m_filename;

		if (strcmp(map1, map2) != 0)
			match = FALSE;

		if (match) {
			// found a match, so return the matlist index
			return i;
		}
	}

	// if we haven't returned yet, then we need to add a new entry and return the index
	matlist.Add(mat);

	return matlist.GetSize()-1;
}

//
// Animation Functions
//

BOOL CHxtApp::ExportSMAnimationSet(float errortolerance, HActionCache *hac, HModelCache *hmc)
{
   FileStream fs, fs2;

   LONG totalitems, curritem = 0;
   LONG numBones = 0;

   FILE * fs3;

   String tempfilename;

   String path = GetLocalRegistryString("AMXtex .X Plugin", "File Path", ".");

   String filename = m_szPath;

   m_hprogressbar = GetProgressBar();
   if (m_hprogressbar) {
      m_hprogressbar->Initialize();
      CString text;
      text.LoadString(IDS_ANIMEXPORTTITLE);
      m_hprogressbar->SetMessage1((LPCTSTR)text);
      m_hprogressbar->SetMessage2((LPCTSTR)filename.GetNameOnly().Get());
   }

   totalitems = 5;	// start process, create channel list, read X file,
					// write SM data, copy/move file

   if (m_hprogressbar) {	// Starting process (this is so it doesn't look like A:M is hung)
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   // Build the channel list for the action
   if (buildChannelList(hac, hmc) == FALSE) {
	   AfxMessageBox("Error creating action channel list", MB_ICONINFORMATION);
	   return false;
   }

   if (m_hprogressbar) {	// Done creating channel list
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   numchannels = m_channellist.GetSize();

   ftmlist.SetSize(ftmlistSize);

   if ( !fs2.Open( m_szPath, HashStream::HS_LOAD ))
      return FALSE;

   // the file load (hash function) breaks the m_hprogressbar (it takes it over during load)
   // and makes it take up the whole bottom of the screen.  We'll delete it here and
   // recreate it in order to get it to work
   if (m_hprogressbar) {
      m_hprogressbar->Uninitialize();
      m_hprogressbar = NULL;
	  }

   m_hprogressbar = GetProgressBar();
   if (m_hprogressbar) {
      m_hprogressbar->Initialize();
      CString text;
      text.LoadString(IDS_ANIMEXPORTTITLE);
      m_hprogressbar->SetMessage1((LPCTSTR)text);
      m_hprogressbar->SetMessage2((LPCTSTR)filename.GetNameOnly().Get());
   }

   GetSMFTMs(fs2); // this function seems to work

   fs2.Close();

   if (m_hprogressbar) {	// done GetSMFTM from X file
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   String fn = m_szPath;

   // open the output file
   tempfilename = fn + "AMXtexTEMP";

   if (!CopyFile(
	   m_szPath,
	   tempfilename.Get(),	// pointer to filename to copy to
	   FALSE)) {		// flag for operation if file exists (overwrite if it exists)
	   AfxMessageBox("AMXtex could not create the temp file for export.", MB_ICONINFORMATION);
	   return FALSE;
   }

   fs3 = fopen (tempfilename.Get(),"at");

   // add a blank line to separate model from AnimationSet
   fprintf(fs3, "\n");

   if (animationname.length() > 0)
	   fprintf(fs3, "AnimationSet %s {\n", animationname.Get());
   else
	   fprintf(fs3, "AnimationSet { \n");

   WriteSMAnimationData(fs3);

   // Close out the AnimationSet
   fprintf(fs3, "}\n");

   fclose (fs3);

   if (m_hprogressbar) {	// Done writing SM animation data
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   // now copy/rename the temp file based on user selection
   if (APONewCopy) { // simply rename the temp file
	   CString newfilename;
	   newfilename = m_szPath;
	   newfilename.Replace(".X", NULL);
	   newfilename.Replace(".x", NULL);

	   newfilename = newfilename + " - " + animationname.Get() + ".X";

	   // delete the new file if it exists because
	   // MoveFile can't move to an existing destination file
	   FILE * stream;

	   if( (stream  = fopen( newfilename, "r" )) != NULL ) {// file exists
		   fclose(stream);
		   CFile::Remove( newfilename ); 
	   }

	   if (!MoveFile(
		   tempfilename.Get(), // pointer to the name of the existing file
		   newfilename       // pointer to the new name for the file
		   )) {
		   AfxMessageBox("AMXtex could not rename the temp file.", MB_ICONINFORMATION);
		   return FALSE;
	   }
   }
   else { // append, so copy it back to the original file
	   if (!CopyFile(
		   tempfilename.Get(),		// pointer to name of an existing file
		   m_szPath,
		   FALSE)) {		// flag for operation if file exists (overwrite if it exists)
		   AfxMessageBox("AMXtex could not overwrite the source file.", MB_ICONINFORMATION);
		   return FALSE;
	   }
	   // don't forget to delete the source file
	   CFile::Remove( tempfilename.Get() );
   }

   if (m_hprogressbar) {	// Done copying/moving file
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }
   
   if (m_hprogressbar) {
      m_hprogressbar->Uninitialize();
      m_hprogressbar = NULL;
	  }


   m_channellist.SetSize(0);
   m_channellist.RemoveAll();
   m_channellist.FreeExtra();
   ftmlist.SetSize(0);
   ftmlist.RemoveAll();
   ftmlist.FreeExtra();

   for (int i=animlist.GetSize()-1;i>=0;i--) {
	   animlist[i].frame.SetSize(0);
	   animlist[i].frame.RemoveAll();
	   animlist[i].frame.FreeExtra();
   }
   animlist.SetSize(0);
   animlist.RemoveAll();
   animlist.FreeExtra();

   return TRUE;
}


BOOL CHxtApp::ExportRMAnimationSet(float errortolerance, HActionCache *hac, HModelCache *hmc)
{
   FileStream fs, fs2; 

   LONG totalitems, curritem = 0;

   FILE * fs3;

   String tempfilename;

   String path = GetLocalRegistryString("AMXtex .X Plugin", "File Path", ".");

   String filename = m_szPath;

   m_hprogressbar = GetProgressBar();
   if (m_hprogressbar) {
      m_hprogressbar->Initialize();
      CString text;
      text.LoadString(IDS_ANIMEXPORTTITLE);
      m_hprogressbar->SetMessage1((LPCTSTR)text);
      m_hprogressbar->SetMessage2((LPCTSTR)filename.GetNameOnly().Get());
   }

   totalitems = 5;	// start process, create channel list, read X file,
					// write RM data, copy/move file

   if (m_hprogressbar) {	// Starting process (this is so it doesn't look like A:M is hung)
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   // Build the channel list for the action
   if (buildChannelList(hac, hmc) == FALSE) {
	   AfxMessageBox("Error creating action channel list", MB_ICONINFORMATION);
	   return false;
   }

   if (m_hprogressbar) {	// Done creating channel list
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   numchannels = m_channellist.GetSize();

   ftmlist.SetSize(ftmlistSize);

   if ( !fs2.Open( m_szPath, HashStream::HS_LOAD ))
      return FALSE;

   // the file load (hash function) breaks the m_hprogressbar (it takes it over during load)
   // and makes it take up the whole bottom of the screen.  We'll delete it here and
   // recreate it in order to get it to work
   if (m_hprogressbar) {
      m_hprogressbar->Uninitialize();
      m_hprogressbar = NULL;
	  }

   m_hprogressbar = GetProgressBar();
   if (m_hprogressbar) {
      m_hprogressbar->Initialize();
      CString text;
      text.LoadString(IDS_ANIMEXPORTTITLE);
      m_hprogressbar->SetMessage1((LPCTSTR)text);
      m_hprogressbar->SetMessage2((LPCTSTR)filename.GetNameOnly().Get());
   }

   GetRMFTMs(fs2);

   fs2.Close();

   if (m_hprogressbar) {	// done GetRMFTM from X file
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   String fn = m_szPath;
   // open the output file
   tempfilename = fn + "AMXtexTEMP";

   if (!CopyFile(
	   m_szPath,
	   tempfilename.Get(),	// pointer to filename to copy to
	   FALSE)) {		// flag for operation if file exists (overwrite if it exists)
	   AfxMessageBox("AMXtex could not create the temp file for export.", MB_ICONINFORMATION);
	   return FALSE;
   }

   fs3 = fopen (tempfilename.Get(),"at");

   if (animationname.length() > 0)
	   fprintf(fs3, "AnimationSet %s {\n", animationname.Get());
   else
	   fprintf(fs3, "AnimationSet { \n");

   WriteRMAnimationData(fs3);

   fprintf(fs3, "}");

   fclose (fs3);

   if (m_hprogressbar) {	// Done writing RM animation data
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }

   // now copy/rename the temp file based on user selection
   if (APONewCopy) { // simply rename the temp file
	   CString newfilename;
	   newfilename = m_szPath;
	   newfilename.Replace(".X", NULL);
	   newfilename.Replace(".x", NULL);

	   newfilename = newfilename + " - " + animationname.Get() + ".X";

	   // delete the new file if it exists
	   // besides, MoveFile can't move to an existing destination file
	   FILE * stream;

	   if( (stream  = fopen( newfilename, "r" )) != NULL ) {// file exists
		   fclose(stream);
		   CFile::Remove( newfilename ); 
	   }

	   if (!MoveFile(
		   tempfilename.Get(), // pointer to the name of the existing file
		   newfilename       // pointer to the new name for the file
		   )) {
		   AfxMessageBox("AMXtex could not rename the temp file.", MB_ICONINFORMATION);
		   return FALSE;
	   }
   }
   else { // append, so copy it back to the original file
	   if (!CopyFile(
		   tempfilename.Get(),		// pointer to name of an existing file
		   m_szPath,
		   FALSE)) {		// flag for operation if file exists (overwrite if it exists)
		   AfxMessageBox("AMXtex could not overwrite the source file.", MB_ICONINFORMATION);
		   return FALSE;
	   }
	   // don't forget to delete the source file
	   CFile::Remove( tempfilename.Get() );
   }

   if (m_hprogressbar) {	// Done copying/moving file
	   curritem++;
	   m_hprogressbar->SetProgress((int)((float)curritem/totalitems*100));
   }
   
   if (m_hprogressbar) {
      m_hprogressbar->Uninitialize();
      m_hprogressbar = NULL;
	  }

   m_channellist.SetSize(0);
   m_channellist.RemoveAll();
   m_channellist.FreeExtra();
   ftmlist.SetSize(0);
   ftmlist.RemoveAll();
   ftmlist.FreeExtra();

   for (int i=animlist.GetSize()-1;i>=0;i--) {
	   animlist[i].frame.SetSize(0);
	   animlist[i].frame.RemoveAll();
	   animlist[i].frame.FreeExtra();
   }
   animlist.SetSize(0);
   animlist.RemoveAll();
   animlist.FreeExtra();

   return TRUE;
}

int CHxtApp::GetChannelCount(const CString &bonename)
{
	BOOL foundmatch=FALSE;
	int i = 0;

	while (!foundmatch && i < numchannels) {
		Channel2 &channel = m_channellist[i];

		if (channel.m_name == bonename)
			foundmatch=TRUE;
		else
			i++;
	}
	
	if (i == numchannels)
		return -1;
	else
		return i;
}

void CHxtApp::WriteSMAnimationData(FILE * fs3)
{
	Matrix44 ftm;
	Matrix34 actionmatrix;
	Matrix44 anim, newanim;
	Matrix44 final;
	Vector actionquatVec;

	int actionindex;
	int x, y;
	Vector4 v1, v2, v3, v4;
	Vector4 v;
	char vstring[80];
	CString vstring2;
	int numberofframes = 0;
	int animlistnum;

	Matrix44 ftmtranslation;
	Matrix44 ftmrotation;
	Vector4 ftmTransVec, ftmTransVec2, ftmTransVecFinal;
	Matrix44 ftmnegativetranslation;

	Matrix44 newTranslation;

	Vector4 rotatedTransVec;


	//first time through, we must figure out how many frames there are
	// do a dummy run through the time sequence -- probably a waste of code;
	// can be optimized later
	
	for (Time time; time<=totaltime2; time.StepFrame(fps)) {
		numberofframes++;
	}

	// create the animlist data now, init to proper valuess
	animlist.SetSize(ftmlistpos);
	for (int i=0;i<ftmlistpos;i++) {
		animlist[i].frame.SetSize(numberofframes);
		// init the animlist frames to identity
		for (int j=0;j<numberofframes;j++) {
			animlist[i].frame[j].SetIdentity();
		}
	}

	for (i=0;i<ftmlistpos;i++) {

		x = 0;
		ftm = ftmlist[i].matrix;

		ftmnegativetranslation.SetIdentity();
		ftmTransVec = ftm[3];
		ftmnegativetranslation[3] = Vector4(-1 * ftmTransVec[0], -1 * ftmTransVec[1], -1 * ftmTransVec[2], ftmTransVec[3]);
			
		ftmrotation = ftm;
		ftmrotation *= ftmnegativetranslation;
		// ftmrotation is now located at the origin

		// Set the translation matrix to identity plus the ftm's translation
		ftmtranslation.SetIdentity();
		ftmtranslation[3] = ftm[3];
		
		animlistnum = ftmlist[i].depth;

		actionindex = GetChannelCount(ftmlist[i].bonename);

		if (actionindex > -1) {
			fprintf(fs3, "\n  Animation {\n");
			fprintf(fs3, "    {");
			fprintf(fs3, "%s", ftmlist[i].bonename);
			fprintf(fs3, "}\n");
			fprintf(fs3, "    AnimationKey {\n");
			fprintf(fs3, "      4;\n");
			fprintf(fs3, "      %d;\n", numberofframes);

			for (Time time; time<=totaltime2; time.StepFrame(fps)) {
				y = x * frameunits; // frameunits is set by user

				fprintf(fs3, "      %d; 16; ", y);

				actionmatrix = GetMatrix(actionindex, time);

				anim = ToMatrix44(actionmatrix);
				anim = anim.Transpose();

				actionquatVec = GetTSQTrans(actionindex, time);

				// Multiply the translation from anim and ftm
				newTranslation.SetIdentity();
				ftmTransVec2 = anim[3];
				ftmnegativetranslation[3] = Vector4(-1 * ftmTransVec2[0], -1 * ftmTransVec2[1], -1 * ftmTransVec2[2], ftmTransVec2[3]);
				anim *= ftmnegativetranslation;

				rotatedTransVec = ftmrotation.Transpose() * ftmTransVec2;

				ftmTransVecFinal[0] = ftmTransVec[0] + rotatedTransVec[0];
				ftmTransVecFinal[1] = ftmTransVec[1] + rotatedTransVec[1];
				ftmTransVecFinal[2] = ftmTransVec[2] + rotatedTransVec[2];

				final = anim * ftmrotation;     // add rotation

				final[3][0] = ftmTransVecFinal[0];
				final[3][1] = ftmTransVecFinal[1];
				final[3][2] = ftmTransVecFinal[2];

				for (int z=0;z<4;z++) {
					v = final[z];

					sprintf(vstring, "%.6f,%.6f,%.6f,%.6f", v[0],v[1],v[2],v[3]);
					vstring2 = vstring;
					fprintf(fs3, "%s", vstring2);
					if (z < 3)
						fprintf(fs3, ",");
					else
						fprintf(fs3, ";;");
				}

				if (x<(numberofframes-1))
					fprintf(fs3, ",\n");
				else
					fprintf(fs3, ";\n");

				x++;
			}
			fprintf(fs3, "    }\n");
			fprintf(fs3, "  }\n");
		}
	}
}

void CHxtApp::WriteRMAnimationData(FILE * fs3)
{
	Matrix44 ftm;
	Matrix34 actionmatrix;
	Matrix44 anim, newanim;
	Matrix44 final;

	Quaternion tsqquat;
	Vector tsqquatvec;
	Vector tsqscalevec;
	float tsqquatW;

	Quaternion actionquat;
	Vector actionquatVec;
	float actionquatW;

	int actionindex;
	int x, y;
	Vector4 v1, v2, v3, v4;
	Vector4 v;
	char vstring[80];
	CString vstring2;
	int numberofframes = 0;
	int animlistnum;

	Matrix44 ftmtranslation;
	Matrix44 ftmrotation;
	Vector4 ftmTransVec;
	Matrix44 ftmnegativetranslation;

	Matrix44 newTranslation;

	Matrix33 mat33;

	BOOL ExportRot, ExportTrans, ExportScale;

	//first time through, we must figure out how many frames there are
	// do a dummy run through the time sequence -- probably a waste of code;
	// can be optimized later
	
	for (Time time; time<=totaltime2; time.StepFrame(fps)) {
		numberofframes++;
	}

	// create the animlist data now, init to proper valuess
	animlist.SetSize(ftmlistpos);
	for (int i=0;i<ftmlistpos;i++) {
		animlist[i].frame.SetSize(numberofframes);
		// init the animlist frames to identity
		for (int j=0;j<numberofframes;j++) {
			animlist[i].frame[j].SetIdentity();
		}
	}

	for (i=0;i<ftmlistpos;i++) {
		x = 0;
		ftm = ftmlist[i].matrix;

		ftmnegativetranslation.SetIdentity();
		ftmTransVec = ftm[3];
		ftmnegativetranslation[3] = Vector4(-1 * ftmTransVec[0], -1 * ftmTransVec[1], -1 * ftmTransVec[2], ftmTransVec[3]);
			
		ftmrotation = ftm;
		ftmrotation *= ftmnegativetranslation;
		// ftmrotation is now located at the origin

		// Set the translation matrix to identity plus the ftm's translation
		ftmtranslation.SetIdentity();
		ftmtranslation[3] = ftm[3];

		ExportRot = FALSE;
		ExportTrans = FALSE;
		ExportScale = FALSE;

		v1 = ftm[0];
		v2 = ftm[1];
		v3 = ftm[2];
		v4 = ftm[3];

		// get the frame scalevec
		tsqscalevec[0] = v1[0];
		tsqscalevec[1] = v2[1];
		tsqscalevec[2] = v3[2];

		animlistnum = ftmlist[i].depth;
		actionindex = GetChannelCount(ftmlist[i].bonename);

		if (actionindex > -1) {

			ExportRot = TRUE;
			ExportTrans = TRUE;
			ExportScale = TRUE;

			if (ExportRot || ExportTrans || ExportScale) {
				// one of them should be exported, so add the animation header
				fprintf(fs3, "\n  Animation {\n");
				fprintf(fs3, "    {%s}\n", ftmlist[i].bonename);
			}

			if (ExportRot) {
				// Write out the rotation data
				fprintf(fs3, "    AnimationKey {\n");
				fprintf(fs3, "      0;\n");

				fprintf(fs3, "      %d;\n", numberofframes);

				for (Time time; time<=totaltime2; time.StepFrame(fps)) {
					y = x * frameunits; // frameunits is set by user

					mat33 = ToMatrix33(ftm);
					actionquat = Quaternion(mat33);
					fprintf(fs3, "      %d; 4; ", y);

					tsqquat = GetTSQRot(actionindex, time);
					tsqquatvec = tsqquat.GetV();	// get the vector
					tsqquatW = tsqquat.GetW();

					Matrix44 amrotmat = ToMatrix44(Matrix34(tsqquat));
					amrotmat = amrotmat.Transpose();

					Matrix44 finMat = amrotmat * ftm;

					Quaternion finquat = Quaternion(ToMatrix34(finMat.Transpose()));

					actionquat = Quaternion(finquat);
					actionquatVec = actionquat.GetV();
					actionquatW = actionquat.GetW();

					sprintf(vstring, "%.6f,%.6f,%.6f,%.6f;;",
						actionquatW,
						actionquatVec.x * -1.0,
						actionquatVec.y * -1.0,
						actionquatVec.z * -1.0 * modelorientation);
					vstring2 = vstring;
					fprintf(fs3, "%s", vstring2);

					if (x<(numberofframes-1))
						fprintf(fs3, ",\n");
					else
						fprintf(fs3, ";\n");

					x++;
				}
				fprintf(fs3, "    }\n");
			}

			if (ExportScale) {

				// Write out the scale data
				x=0;
				fprintf(fs3, "    AnimationKey {\n");
				fprintf(fs3, "      1;\n");
				fprintf(fs3, "      %d;\n", numberofframes);

				for (Time time2; time2<=totaltime2; time2.StepFrame(fps)) {
					y = x * frameunits; // frameunits is set by user

					fprintf(fs3, "      %d; 3; ", y);

					actionquatVec = GetTSQScale(actionindex, time2);

					sprintf(vstring, "%.6f,%.6f,%.6f;;",
						actionquatVec.x, //* tsqscalevec[0],
						actionquatVec.y, //* tsqscalevec[1],
						actionquatVec.z * modelorientation);// * tsqscalevec[2]);
					vstring2 = vstring;

					fprintf(fs3, "%s", vstring2);

					if (x<(numberofframes-1))
						fprintf(fs3, ",\n");
					else
						fprintf(fs3, ";\n");

					x++;
				}
				fprintf(fs3, "    }\n");
			}

			if (ExportTrans) {

				// Write out the translation data
				x=0;
				fprintf(fs3, "    AnimationKey {\n");
				fprintf(fs3, "      2;\n");
				fprintf(fs3, "      %d;\n", numberofframes);

				for (Time time3; time3<=totaltime2; time3.StepFrame(fps)) {
					y = x * frameunits; // frameunits is set by user

					fprintf(fs3, "      %d; 3; ", y);

					actionquatVec = GetTSQTrans(actionindex, time3);

					actionquatVec = ftmrotation.Transpose() * actionquatVec;

					sprintf(vstring, "%.6f,%.6f,%.6f;;",
						actionquatVec.x + ftmTransVec[0],
						actionquatVec.y + ftmTransVec[1],
						(actionquatVec.z * modelorientation) + ftmTransVec[2]);
					vstring2 = vstring;

					fprintf(fs3, "%s", vstring2);

					if (x<(numberofframes-1))
						fprintf(fs3, ",\n");
					else
						fprintf(fs3, ";\n");

					x++;
				}
				fprintf(fs3, "    }\n");
			}

			if (ExportRot || ExportTrans || ExportScale)
				fprintf(fs3, "  }\n");	// close the animation
		}
	}
}

void CHxtApp::GetSMFTMs(FileStream &fs2)
{
	// Create a list of bone names and ftms in exportclasses.h
	// bone name is used for searching for ftms down the line
	// The final animation output will use this list for the
	// frame multiplication.
	String line;
	char *charline;
	CString linedata;
	BOOL gotline = FALSE;
	BOOL frameparse = TRUE;
	BOOL frametransform = FALSE;
	int findmesh, findframe, framepos;
	int boneorderpos = 0;
	int x;
	CString BoneName;
	int bonenamestart;
	int bonenameend;
	Vector4 v[4];
	float f1,f2,f3,f4;
	Matrix44 testmatrix;
	char vstring[40];
	Vector4 v1,v2,v3,v4;

	while (frameparse) {
		gotline = fs2.ReadLine(line);

		if (gotline) {
			charline = line.Get();
			linedata = charline;
			findmesh = linedata.Find("Mesh ");
			findframe = linedata.Find("Frame ");
			if (findmesh > -1)
				frameparse = FALSE;

			if ((findmesh == -1) &&	// is NOT a mesh line
				(findframe > -1))	// IS a frame line
			{
				// find the bone name
				framepos = findframe/2; // halve the # of spaces in front of "Frame"
				framepos++;				// and add 1

				bonenamestart = linedata.Find("Frame");
				bonenamestart = bonenamestart + 6;
				bonenameend = linedata.ReverseFind('{');
				bonenameend = bonenameend - 1;
				bonenameend = bonenameend - bonenamestart;

				BoneName = linedata.Mid(bonenamestart,bonenameend);
				ftmlist[ftmlistpos].bonename = BoneName;

				// read the frame transform line (no use;trash it)
				gotline = fs2.ReadLine(line); // read another line
				
				for (x=0;x<4;x++) {
					gotline = fs2.ReadLine(line);
					charline = line.Get();
					linedata = charline;
					linedata.Replace(',', ' ');  // replace commas
					linedata.Replace(';', ' ');  // replace semicolons
					sscanf(linedata, "%f %f %f %f", &f1, &f2, &f3, &f4);

					sprintf(vstring,"%.6f %.6f %.6f %.6f",f1,f2,f3,f4);

					v[x].Set(f1,f2,f3,f4); 
				}

				ftmlist[ftmlistpos].matrix = Matrix44(v[0],v[1],v[2],v[3]);
				ftmlist[ftmlistpos].depth = framepos;
				ftmlistpos++;

				// Check if the ftmlist has outgrown the ftmlist size and grow if needed
				if (ftmlistpos == ftmlistSize) {
					// grow the ftmlist array by 32 (to prevent too many grows)
					ftmlistSize += 32;
					ftmlist.SetSize(ftmlistSize);
				}
			}
		}
	}

} // End CalculateAnimationMatrices

void CHxtApp::GetRMFTMs(FileStream &fs2)
{
	// Create a list of bone names and ftms in exportclasses.h
	// bone name is used for searching for ftms down the line
	// The final animation output will use this list for the
	// frame multiplication.
	String line;
	char *charline;
	CString linedata;
	BOOL gotline = FALSE;
	BOOL frameparse = TRUE;
	BOOL frametransform = FALSE;
	int findmesh, findframe, framepos;
	int boneorderpos = 0;
	int x;
	CString BoneName;
	int bonenamestart;
	int bonenameend;
	Vector4 v[4];
	float f1,f2,f3,f4;
	Matrix44 testmatrix;
	char vstring[40];
	Vector4 v1,v2,v3,v4;

	while (frameparse) {
		gotline = fs2.ReadLine(line);

		if (gotline) {
			charline = line.Get();
			linedata = charline;
			findmesh = linedata.Find("End RootFrame");
			findframe = linedata.Find("Frame ");
			if (findmesh > -1)
				frameparse = FALSE;

			if ((findmesh == -1) &&	// is NOT a mesh line
				(findframe > -1))	// IS a frame line
			{
				// find the bone name
				framepos = findframe/2; // halve the # of spaces in front of "Frame"
				framepos++;				// and add 1

				bonenamestart = linedata.Find("Frame");
				bonenamestart = bonenamestart + 6;
				bonenameend = linedata.ReverseFind('{');
				bonenameend = bonenameend - 1;
				bonenameend = bonenameend - bonenamestart;

				BoneName = linedata.Mid(bonenamestart,bonenameend);
				ftmlist[ftmlistpos].bonename = BoneName;

				// read the frame transform line (no use;trash it)
				gotline = fs2.ReadLine(line); // read another line
				
				for (x=0;x<4;x++) {
					gotline = fs2.ReadLine(line);
					charline = line.Get();
					linedata = charline;
					linedata.Replace(',', ' ');  // replace commas
					linedata.Replace(';', ' ');  // replace semicolons
					sscanf(linedata, "%f %f %f %f", &f1, &f2, &f3, &f4);

					sprintf(vstring,"%.6f %.6f %.6f %.6f",f1,f2,f3,f4);

					v[x].Set(f1,f2,f3,f4); 
				}

				ftmlist[ftmlistpos].matrix = Matrix44(v[0],v[1],v[2],v[3]);
				ftmlist[ftmlistpos].depth = framepos;
				ftmlistpos++;

				// Check if the ftmlist has outgrown the ftmlist size and grow if needed
				if (ftmlistpos == ftmlistSize) {
					// grow the ftmlist array by 32 (to prevent too many grows)
					ftmlistSize += 32;
					ftmlist.SetSize(ftmlistSize);
				}
			}
		}
	}
} // End CalculateAnimationMatrices


Matrix34 CHxtApp::CreateMatrix( const TSQ &tsq )
{

	// modify these vars based on right/left handed-ness
	// scale needs no modification
	Vector tsqtranslate = tsq.translate;
	
	// modify the trans based on orientation
	tsqtranslate.z = tsqtranslate.z * modelorientation;
	
	Matrix34 actionmatrix( tsq.quat );
	MatrixTimesScale( actionmatrix, tsq.scale );
	actionmatrix.SetTranslate( tsqtranslate );

	return actionmatrix;
}

Matrix34 CHxtApp::GetMatrix(int actionindex, Time time)
{
   TSQ tsq = m_channellist[actionindex].GetTSQ(time);

   return CreateMatrix(tsq);
}


Quaternion CHxtApp::GetTSQRot(int actionindex, Time time)
{
   TSQ tsq = m_channellist[actionindex].GetTSQ(time);

   return tsq.quat;
}

Vector CHxtApp::GetTSQTrans(int actionindex, Time time)
{
   TSQ tsq = m_channellist[actionindex].GetTSQ(time);

   return tsq.translate;
}

Vector CHxtApp::GetTSQScale(int actionindex, Time time)
{
   TSQ tsq = m_channellist[actionindex].GetTSQ(time);

   return tsq.scale;
}

// Channel List Code (modified from Hash's KFM code example)

bool CHxtApp::buildChannelList(HActionCache *hac, HModelCache *hmc) {
	
	float errortolerance = 1.0f;

	// confirm that we have both a model and an action to work with
	if ((hmc == NULL) ||
		(hac == NULL))
		return false;
	else {
		// start by removing any existing channels, as this can be called
		// multiple times during export
		m_channellist.SetSize(0);
		m_channellist.FreeExtra();
		m_channellist.RemoveAll();

		// now, populate the channel list using our scenemodel and sceneaction
		
		HModel *hm = hac->EditWithModelCache(hmc);
		HActionCache *hunconstrainedac = hac->CreateUncontrainedActionCache(hm, errortolerance);
		float scalefactor = hunconstrainedac->GetScaleFactor();
		hac->EndEditWithModelCache(hm);
		if (hunconstrainedac==NULL)
			return false;

		hm = hunconstrainedac->EditWithModelCache(hmc);

		totaltime2 = hunconstrainedac->GetTotalTime().GetSeconds();

		// Count bone shortcuts, these are bones that have drivers
		HAnimObjectShortcut *bonesfolder = (HAnimObjectShortcut *) hunconstrainedac->FindChildByName("Bones");// Finds "Bones" shortcut under driver folder

		WORD count = 1;
		if (bonesfolder) {
			for (HAnimObjectShortcut *boneshortcut = (HAnimObjectShortcut *)bonesfolder->GetChild(); boneshortcut; boneshortcut=(HAnimObjectShortcut *)boneshortcut->GetSibling(), count++);
		}

		m_channellist.SetSize(count);

		int cCount = 0;

		// model bone
		SaveBoneTransformDriver(cCount, hm, NULL, errortolerance, scalefactor);

		if (bonesfolder) {
			for (HAnimObjectShortcut *boneshortcut = (HAnimObjectShortcut *)bonesfolder->GetChild(); boneshortcut; boneshortcut=(HAnimObjectShortcut *)boneshortcut->GetSibling()) {
				cCount++;

				HBone *bone = (HBone *)boneshortcut->GetAnimObject();

				SaveBoneTransformDriver(cCount, bone, boneshortcut, errortolerance, scalefactor);
			}
		}

		hunconstrainedac->EndEditWithModelCache(hm);
		HTreeObject::DeleteHandle(hunconstrainedac);

		return true;
	}
}

HDriverShortcut *FindDriverShortcut( HBone *hbone, HProperty *hproperty ) {
	for (HPropertyDriver *driver=hbone->GetTransform()->GetHeadDriver(); driver; driver=(HPropertyDriver *)driver->GetNextDriver()) // might should be GetNextDriver()
		if (driver->GetProperty() == hproperty && driver->IsSimpleDriverShortcut())
			return (HDriverShortcut *)driver;
		return NULL;
}

void CHxtApp::SaveBoneTransformDriver(int channelcount, HBone *bone, HAnimObjectShortcut *boneshortcut, float errortolerance, float scalefactor) {
	String name;

	Channel2 &c2data = m_channellist[channelcount];

	if (bone) {
		name = bone->GetName();
		c2data.m_name = name.Get();
		c2data.m_name = GetValidDirectXName(c2data.m_name);
	}
	else if (boneshortcut) {
		name = boneshortcut->GetName();
		c2data.m_name = name.Get();
		c2data.m_name = GetValidDirectXName(c2data.m_name);
	}
	else
		return;

	Vector bonedirection, rolldirection, parentbonedirection, parentrolldirection;
	HRotateDriver *rotatedriver = NULL;
	HChannelDriver *translatedriver[3];
	HChannelDriver *scaledriver[3];
	
	for (int i=0; i < 3; i++) {
		translatedriver[i] = NULL;
		scaledriver[i] = NULL;
	}

	if (bone) {
		bone->ComputeBoneDirections(bonedirection, rolldirection, parentbonedirection, parentrolldirection);
		//quatdriver->ComputeBoneToParent(bonedirection, rolldirection, parentbonedirection, parentrolldirection);

		HDriverShortcut *drivershortcut;

		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetRotate() );
		if (drivershortcut)
			rotatedriver = (HRotateDriver *)drivershortcut->GetDriver();

		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetTranslate()->GetX() );
		if (drivershortcut)
			translatedriver[0] = (HChannelDriver *)drivershortcut->GetDriver(); // could be a bad assumption that these are HChannelDrivers in the future, they can be any derivitive of HFloatDriver
		
		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetTranslate()->GetY() );
		if (drivershortcut)
			translatedriver[1] = (HChannelDriver *)drivershortcut->GetDriver();

		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetTranslate()->GetZ() );
		if (drivershortcut)
			translatedriver[2] = (HChannelDriver *)drivershortcut->GetDriver();

		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetScale()->GetX());
		if (drivershortcut)
			scaledriver[0] = (HChannelDriver *)drivershortcut->GetDriver();

		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetScale()->GetY());
		if (drivershortcut)
			scaledriver[1] = (HChannelDriver *)drivershortcut->GetDriver();
		
		drivershortcut = FindDriverShortcut( bone, bone->GetTransform()->GetScale()->GetZ());
		if (drivershortcut)
			scaledriver[2] = (HChannelDriver *)drivershortcut->GetDriver();
	}
	else { //This action has drivers for bones that are not in our default model
		   //Lets just find the drivers ourselves
		
		// could be a bad assumption that these are HChannelDrivers in the future, they can be any derivitive of HFloatDriver
		HPropertyShortcut *transform = (HPropertyShortcut *)boneshortcut->FindChildByName("Transform");// just a place holder
		if (transform) {
			rotatedriver = (HRotateDriver *)transform->FindChildByName("Rotate");

			HPropertyShortcut *translatefolder = (HPropertyShortcut *)transform->FindChildByName("Translate");
			if (translatefolder) {
				translatedriver[0] = (HChannelDriver *)translatefolder->FindChildByName("X");
				translatedriver[1] = (HChannelDriver *)translatefolder->FindChildByName("Y");
				translatedriver[2] = (HChannelDriver *)translatefolder->FindChildByName("Z");
			}

			HPropertyShortcut *scalefolder = (HPropertyShortcut *)transform->FindChildByName("Scale");
			if (scalefolder) {
				scaledriver[0] = (HChannelDriver *)scalefolder->FindChildByName("X");
				scaledriver[1] = (HChannelDriver *)scalefolder->FindChildByName("Y");
				scaledriver[2] = (HChannelDriver *)scalefolder->FindChildByName("Z");
			}
		}

		if (rotatedriver)
			rotatedriver->ComputeBoneDirections(bonedirection, rolldirection, parentbonedirection, parentrolldirection);
		else {
			bonedirection.Set(0.0f,0.0f,1.0f);
			rolldirection.Set(0.0f,1.0f,0.0f);
			parentbonedirection.Set(0.0f,0.0f,1.0f);
			parentrolldirection.Set(0.0f,1.0f,0.0f);
		}
	}

	c2data.ComputeParentToBone(bonedirection, rolldirection, parentbonedirection, parentrolldirection);

	HQuaternionRotateDriver *quatrotatedriver = NULL;

	if (rotatedriver && rotatedriver->GetObjectType()!=HOT_QUATERNIONROTATEDRIVER) {
		// Must be EULER or Vector, Lets build our own Quaternion based rotate driver, 
		// and jam keyframes in on every frame and then reduce.  This will get us the same motion
		quatrotatedriver = HQuaternionRotateDriver::New(rotatedriver->GetProperty());
		Time totaltime = rotatedriver->GetLastKey();

		Time prevtime, time;
		for (time=0; time <= totaltime; time.StepFrame(GetHProject()->GetFPS())) {
			TimeRange tr = FOREVER;
			Quaternion quat = rotatedriver->GetValue(time, tr);
			quatrotatedriver->StoreValue(time, quat);
		}
		rotatedriver = quatrotatedriver;

		if (errortolerance)
			quatrotatedriver->Reduce(errortolerance);
	}

	SaveKeyFrames(channelcount, (HQuaternionRotateDriver *)rotatedriver, translatedriver, scaledriver, scalefactor);
	if (quatrotatedriver)
		delete quatrotatedriver;
}

enum { HOLD, LINEAR, SPLINE };

BOOL CHxtApp::SaveKeyFrames(int channelcount, HQuaternionRotateDriver *rotatedriver, HChannelDriver *translatedriver[], HChannelDriver *scaledriver[], float scalefactor) {
	union {
		BYTE flags;
		struct {
			BYTE defaulttangents : 1,
				 interpmethod    : 2;
		};
	};

	Channel2 &c2data = m_channellist[channelcount];

	int chanitemindex = 0;

	// save translates
	for (int i=0; i < 3; i++) {
		HChannelDriver *channeldriver = translatedriver[i];
		if (channeldriver && channeldriver->CountKeys())
			SaveChannelDriver(channelcount, chanitemindex, channeldriver, 1.0f/scalefactor);
		else {
			//AfxMessageBox("Setting a single translation keyframe", MB_ICONINFORMATION);
			//fs << (WORD)1 << 0.0f << 0.0f << flags;
			//       numkf     time    value   flags
			flags = 0;
			interpmethod = SPLINE;
			defaulttangents = TRUE;
			Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];
			chanitem->m_keyframes.SetSize(1);
			Keyframe &kf = chanitem->m_keyframes[0];
			kf.m_time.SetSeconds(0.0f);
			kf.m_value = 0.0f;
			kf.m_flags = flags;
		}

		chanitemindex++;
	}

	chanitemindex = 3;

	// save scales
	for (i=0; i < 3; i++) {
		HChannelDriver *channeldriver = scaledriver[i];
		if (channeldriver && channeldriver->CountKeys())
			SaveChannelDriver(channelcount, chanitemindex, channeldriver, 100.0f);
		else {
			//AfxMessageBox("Setting a single scale keyframe", MB_ICONINFORMATION);
			//fs << (WORD)1 << 0.0f << 1.0f << flags;
			//       numkf     time    value   flags
			flags = 0;
			interpmethod = SPLINE;
			defaulttangents = TRUE;
			Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];
			chanitem->m_keyframes.SetSize(1);
			Keyframe &kf = chanitem->m_keyframes[0];
			kf.m_time.SetSeconds(0.0f);
			kf.m_value = 1.0f;
			kf.m_flags = flags;
		}

		chanitemindex++;
	}

	chanitemindex = 6;

	// save quaternion rotate
	if (rotatedriver) {
		flags = 0;
		interpmethod = SPLINE;
		defaulttangents = TRUE;

		// could be a bad assumption that these are HChannelDrivers in the future
		HChannelDriver *channeldriver = (HChannelDriver *)rotatedriver->GetX();
		if (channeldriver && channeldriver->CountKeys())
			SaveChannelDriver(channelcount, chanitemindex, (HChannelDriver *)channeldriver, 100.0f);
		else {
			//AfxMessageBox("Setting a single X rot keyframe", MB_ICONINFORMATION);
			//fs << (WORD)1 << 0.0f << 0.0f << flags;//x
			//       numkf     time    value   flags
			Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];
			chanitem->m_keyframes.SetSize(1);
			Keyframe &kf = chanitem->m_keyframes[0];
			kf.m_time.SetSeconds(0.0f);
			kf.m_value = 0.0f;
			kf.m_flags = flags;
		}

		chanitemindex++;

		channeldriver = (HChannelDriver *)rotatedriver->GetY();
		if (channeldriver && channeldriver->CountKeys())
			SaveChannelDriver(channelcount, chanitemindex, (HChannelDriver *)channeldriver, 100.0f);
		else {
			//AfxMessageBox("Setting a single Y rot keyframe", MB_ICONINFORMATION);
			//fs << (WORD)1 << 0.0f << 0.0f << flags;//y
			//       numkf     time    value   flags
			Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];
			chanitem->m_keyframes.SetSize(1);
			Keyframe &kf = chanitem->m_keyframes[0];
			kf.m_time.SetSeconds(0.0f);
			kf.m_value = 0.0f;
			kf.m_flags = flags;
		}

		chanitemindex++;

		channeldriver = (HChannelDriver *)rotatedriver->GetZ();
		if (channeldriver && channeldriver->CountKeys())
			SaveChannelDriver(channelcount, chanitemindex, (HChannelDriver *)channeldriver, 100.0f);
		else {
			//AfxMessageBox("Setting a single Z rot keyframe", MB_ICONINFORMATION);
			//fs << (WORD)1 << 0.0f << 0.0f << flags;//z
			//       numkf     time    value   flags
			Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];
			chanitem->m_keyframes.SetSize(1);
			Keyframe &kf = chanitem->m_keyframes[0];
			kf.m_time.SetSeconds(0.0f);
			kf.m_value = 0.0f;
			kf.m_flags = flags;
		}

		chanitemindex++;
		
		channeldriver = (HChannelDriver *)rotatedriver->GetW();
		if (channeldriver && channeldriver->CountKeys())
			SaveChannelDriver(channelcount, chanitemindex, (HChannelDriver *)channeldriver, 100.0f);
		else {
			//AfxMessageBox("Setting a single W rot keyframe", MB_ICONINFORMATION);
			//fs << (WORD)1 << 0.0f << 1.0f << flags;//w
			//       numkf     time    value   flags
			Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];
			chanitem->m_keyframes.SetSize(1);
			Keyframe &kf = chanitem->m_keyframes[0];
			kf.m_time.SetSeconds(0.0f);
			kf.m_value = 1.0f;
			kf.m_flags = flags;
		}
	}
	else {
		flags = 0;
		interpmethod = SPLINE;
		defaulttangents = TRUE;
		//fs << (WORD)1 << 0.0f << 0.0f << flags;//x
		//fs << (WORD)1 << 0.0f << 0.0f << flags;//y
		//fs << (WORD)1 << 0.0f << 0.0f << flags;//z
		//fs << (WORD)1 << 0.0f << 1.0f << flags;//w
		Channel2Item *chanitem6 = &c2data.m_channel2item[6];
		chanitem6->m_keyframes.SetSize(1);
		Keyframe &kf6 = chanitem6->m_keyframes[0];
		kf6.m_time.SetSeconds(0.0f);
		kf6.m_value = 0.0f;
		kf6.m_flags = flags;

		Channel2Item *chanitem7 = &c2data.m_channel2item[7];
		chanitem7->m_keyframes.SetSize(1);
		Keyframe &kf7 = chanitem7->m_keyframes[0];
		kf7.m_time.SetSeconds(0.0f);
		kf7.m_value = 0.0f;
		kf7.m_flags = flags;

		Channel2Item *chanitem8 = &c2data.m_channel2item[8];
		chanitem8->m_keyframes.SetSize(1);
		Keyframe &kf8 = chanitem8->m_keyframes[0];
		kf8.m_time.SetSeconds(0.0f);
		kf8.m_value = 0.0f;
		kf8.m_flags = flags;

		Channel2Item *chanitem9 = &c2data.m_channel2item[9];
		chanitem9->m_keyframes.SetSize(1);
		Keyframe &kf9 = chanitem9->m_keyframes[0];
		kf9.m_time.SetSeconds(0.0f);
		kf9.m_value = 1.0f;
		kf9.m_flags = flags;
	}

	return TRUE;
}

void CHxtApp::SaveChannelDriver(int channelcount, int chanitemindex, HChannelDriver *channeldriver, float unit) {
	union {
		BYTE flags;
		struct {
			BYTE defaulttangents : 1,
				 interpmethod    : 2;
		};
	};

	Channel2 &c2data = m_channellist[channelcount];
	Channel2Item *chanitem = &c2data.m_channel2item[chanitemindex];

	int keys = channeldriver->CountKeys();
	chanitem->m_keyframes.SetSize(keys);

	if (keys == 0)
		AfxMessageBox("POSSIBLE ERROR: Setting a NULL keyframe", MB_ICONINFORMATION);

	int kfindex=0;

	for (HChannelCP *cp = channeldriver->GetHeadCP(); cp; cp=cp->GetNext()) {

		Keyframe &kf = chanitem->m_keyframes[kfindex];

		flags = 0;
		CPInterpolationMethod cpmethod = cp->GetInterpolationMethod();
		if (cpmethod == CPI_DEFAULT)
			cpmethod = channeldriver->GetInterpolationMethod();

		switch (cpmethod) {
		   case CPI_HOLD:
			   interpmethod = HOLD;
			   break;
		   case CPI_LINEAR:
			   interpmethod = LINEAR;
			   break;
		   case CPI_SPLINE:
		   case CPI_ZEROSLOPE:
			   interpmethod = SPLINE; // we don't store zero slope, we just write out the tangents
			   break;
		}

		BOOL isdefaulttangents = cp->GetInGamma() == 0.0f;
		isdefaulttangents = isdefaulttangents && cp->GetInMagnitude() == 1.0f;
		isdefaulttangents = isdefaulttangents && cp->GetOutGamma() == 0.0f;
		isdefaulttangents = isdefaulttangents && cp->GetOutMagnitude() == 1.0f;
		isdefaulttangents = isdefaulttangents && cp->IsSmooth();
		isdefaulttangents = isdefaulttangents && cpmethod != CPI_ZEROSLOPE;

		defaulttangents = isdefaulttangents;

		float value = cp->GetValue()/unit;
		kf.m_time.SetSeconds(cp->GetTime().GetSeconds());
		kf.m_value = value;
		kf.m_flags = flags;

		if (!defaulttangents) {
			Vector2 in = cp->GetInTangent();
			in.y /= unit;
			Vector2 out = cp->GetOutTangent();
			out.y /= unit;
			//fs << in << out;
			kf.m_it = in;
			kf.m_ot = out;
		}

		kfindex++;
	}

	for (kfindex=0; kfindex < keys; kfindex++) {
		Keyframe &kf = chanitem->m_keyframes[kfindex];
		if (kf.m_defaulttangents)
			chanitem->CalcTangents(kfindex);
	}
}

// Other Helper Functions

CString CHxtApp::GetValidDirectXName(CString inputName) {

	// this makes the same changes that the MS 3DS X exporter does
	// (search ExportXFile.cpp for "CreateNiceString" function)

	CString validName = inputName;

	// Check for leading number (directx needs letter first)
	char firstChar = validName.GetAt(0);
	if (isdigit(firstChar)) {
		validName.Insert(0,'_');
	}

	int sLength = validName.GetLength();

	// loop through the string and replace any invalid chars
	for (int i=0;i<sLength;i++) {
		char stringChar = validName.GetAt(i);
		// check if it's either an underscore or alphanumeric character
		// note: initially, this is not supporting unicode chars
		if ((stringChar != '_') && !isalnum((unsigned char)stringChar)) // it's not a valid char
			validName.SetAt(i,'_');
	}

	return validName;
}

void CHxtApp::DuplicateNameCheck() {
	// Only cycle through bone names (not channel names, as they should be identical)

	char currentBone[256];

	for (int i=0;i<bonenumber;i++) {
		strcpy(currentBone,bonelist[i].name);
		for (int j=i+1; j<bonenumber;j++) { // start comparing at one higher than current
			if (strcmp(currentBone,bonelist[j].name) == 0) {
				// A duplicate name was found, so export an error
				char debugMsg[512];
				sprintf(debugMsg,"ERROR: Conversion to valid DirectX object/bone names resulted in\nduplicate names.  Please correct before proceeding.\n\nDuplicate name found: %s",currentBone);
				AfxMessageBox(debugMsg, MB_ICONINFORMATION);
			}
		}
	}
}
